import warnings

import dask
import numpy as np
import xarray as xr

from climpred.constants import CLIMPRED_DIMS, CONCAT_KWARGS, PM_CALENDAR_STR

from .checks import (
    has_dims,
    has_valid_lead_units,
    warn_if_chunking_would_increase_performance,
)
from .comparisons import (
    ALL_COMPARISONS,
    COMPARISON_ALIASES,
    HINDCAST_COMPARISONS,
    __m2o,
)
from .exceptions import KeywordError
from .metrics import ALL_METRICS, METRIC_ALIASES
from .prediction import compute_hindcast, compute_perfect_model
from .reference import compute_climatology, compute_persistence
from .stats import dpp

try:
    from .stats import varweighted_mean_period
except ImportError:
    varweighted_mean_period = None
from .utils import (
    _transpose_and_rechunk_to,
    assign_attrs,
    convert_time_index,
    find_start_dates_for_given_init,
    get_comparison_class,
    get_lead_cftime_shift_args,
    get_metric_class,
    lead_units_equal_control_time_stride,
    rechunk_to_single_chunk_if_more_than_one_chunk_along_dim,
    shift_cftime_singular,
)


def _resample(hind, resample_dim):
    """Resample with replacement in dimension ``resample_dim``.

    Args:
        hind (xr.object): input xr.object to be resampled.
        resample_dim (str): dimension to resample along.

    Returns:
        xr.object: resampled along ``resample_dim``.

    """
    to_be_resampled = hind[resample_dim].values
    smp = np.random.choice(to_be_resampled, len(to_be_resampled))
    smp_hind = hind.sel({resample_dim: smp})
    # ignore because then inits should keep their labels
    if resample_dim != "init":
        smp_hind[resample_dim] = hind[resample_dim].values
    return smp_hind


def _resample_iterations(init, iterations, dim="member", dim_max=None, replace=True):
    """Resample over ``dim`` by index ``iterations`` times.

    .. note::
        This gives the same result as `_resample_iterations_idx`. When using dask, the
        number of tasks in `_resample_iterations` will scale with iterations but
        constant chunksize, whereas the tasks in `_resample_iterations_idx` will stay
        constant with increasing chunksize.

    Args:
        init (xr.DataArray, xr.Dataset): Initialized prediction ensemble.
        iterations (int): Number of bootstrapping iterations.
        dim (str): Dimension name to bootstrap over. Defaults to ``'member'``.
        dim_max (int): Number of items to select in `dim`.
        replace (bool): Bootstrapping with or without replacement. Defaults to ``True``.

    Returns:
        xr.DataArray, xr.Dataset: Bootstrapped data with additional dim ```iteration```

    """
    if dim_max is not None and dim_max <= init[dim].size:
        # select only dim_max items
        select_dim_items = dim_max
        new_dim = init[dim].isel({dim: slice(None, dim_max)})
    else:
        select_dim_items = init[dim].size
        new_dim = init[dim]

    if replace:
        idx = np.random.randint(0, init[dim].size, (iterations, select_dim_items))
    elif not replace:
        # create 2d np.arange()
        idx = np.linspace(
            (np.arange(select_dim_items)),
            (np.arange(select_dim_items)),
            iterations,
            dtype="int",
        )
        # shuffle each line
        for ndx in np.arange(iterations):
            np.random.shuffle(idx[ndx])
    idx_da = xr.DataArray(
        idx,
        dims=("iteration", dim),
        coords=({"iteration": range(iterations), dim: new_dim}),
    )
    init_smp = []
    for i in np.arange(iterations):
        idx = idx_da.sel(iteration=i).data
        init_smp2 = init.isel({dim: idx}).assign_coords({dim: new_dim})
        init_smp.append(init_smp2)
    init_smp = xr.concat(init_smp, dim="iteration", **CONCAT_KWARGS)
    init_smp["iteration"] = np.arange(1, 1 + iterations)
    return init_smp


def _resample_iterations_idx(
    init, iterations, dim="member", replace=True, chunk=True, dim_max=None
):
    """Resample over ``dim`` by index ``iterations`` times.

    .. note::
        This is a much faster way to bootstrap than resampling each iteration
        individually and applying the function to it. However, this will create a
        DataArray with dimension ``iteration`` of size ``iterations``. It is probably
        best to do this out-of-memory with ``dask`` if you are doing a large number
        of iterations or using spatial output (i.e., not time series data).

    Args:
        init (xr.DataArray, xr.Dataset): Initialized prediction ensemble.
        iterations (int): Number of bootstrapping iterations.
        dim (str): Dimension name to bootstrap over. Defaults to ``'member'``.
        replace (bool): Bootstrapping with or without replacement. Defaults to ``True``.
        chunk: (bool): Auto-chunk along chunking_dims to get optimal blocksize
        dim_max (int): Number of indices from `dim` to return. Not implemented.

    Returns:
        xr.DataArray, xr.Dataset: Bootstrapped data with additional dim ```iteration```

    """
    if dask.is_dask_collection(init):
        init = init.chunk({"lead": -1, "member": -1})
        init = init.copy(deep=True)

    def select_bootstrap_indices_ufunc(x, idx):
        """Selects multi-level indices ``idx`` from xarray object ``x`` for all
        iterations."""
        # `apply_ufunc` sometimes adds a singleton dimension on the end, so we squeeze
        # it out here. This leverages multi-level indexing from numpy, so we can
        # select a different set of, e.g., ensemble members for each iteration and
        # construct one large DataArray with ``iterations`` as a dimension.
        return np.moveaxis(x.squeeze()[idx.squeeze().transpose()], 0, -1)

    if dask.is_dask_collection(init):
        if chunk:
            chunking_dims = [d for d in init.dims if d not in CLIMPRED_DIMS]
            init = _chunk_before_resample_iterations_idx(
                init, iterations, chunking_dims
            )

    # resample with or without replacement
    if replace:
        idx = np.random.randint(0, init[dim].size, (iterations, init[dim].size))
    elif not replace:
        # create 2d np.arange()
        idx = np.linspace(
            (np.arange(init[dim].size)),
            (np.arange(init[dim].size)),
            iterations,
            dtype="int",
        )
        # shuffle each line
        for ndx in np.arange(iterations):
            np.random.shuffle(idx[ndx])
    idx_da = xr.DataArray(
        idx,
        dims=("iteration", dim),
        coords=({"iteration": range(iterations), dim: init[dim]}),
    )
    transpose_kwargs = (
        {"transpose_coords": False} if isinstance(init, xr.DataArray) else {}
    )
    return xr.apply_ufunc(
        select_bootstrap_indices_ufunc,
        init.transpose(dim, ..., **transpose_kwargs),
        idx_da,
        dask="parallelized",
        output_dtypes=[float],
    )


def _distribution_to_ci(ds, ci_low, ci_high, dim="iteration"):
    """Get confidence intervals from bootstrapped distribution.

    Needed for bootstrapping confidence intervals and p_values of a metric.

    Args:
        ds (xarray object): distribution.
        ci_low (float): low confidence interval.
        ci_high (float): high confidence interval.
        dim (str): dimension to apply xr.quantile to. Default: 'iteration'

    Returns:
        uninit_hind (xarray object): uninitialize hindcast with hind.coords.
    """
    ds = rechunk_to_single_chunk_if_more_than_one_chunk_along_dim(ds, dim)
    if isinstance(ds, xr.Dataset):
        for v in ds.data_vars:
            if np.issubdtype(ds[v].dtype, np.bool_):
                ds[v] = ds[v].astype(np.float_)  # fails on py>36 if boolean dtype
    else:
        if np.issubdtype(ds.dtype, np.bool_):
            ds = ds.astype(np.float_)  # fails on py>36 if boolean dtype
    return ds.quantile(q=[ci_low, ci_high], dim=dim, skipna=False)


def _pvalue_from_distributions(ref_skill, init_skill, metric=None):
    """Get probability that skill of a reference forecast (e.g., persistence or
    uninitialized skill) is larger than initialized skill.

    Needed for bootstrapping confidence intervals and p_values of a metric in
    the hindcast framework. Checks whether a simple forecast like persistence, climatology
    or uninitialized performs better than initialized forecast. Need to keep in
    mind the orientation of metric (whether larger values are better or worse
    than smaller ones.)

    Args:
        ref_skill (xarray object): persistence or uninitialized skill.
        init_skill (xarray object): initialized skill.
        metric (Metric): metric class Metric

    Returns:
        pv (xarray object): probability that simple forecast performs better
                            than initialized forecast.
    """
    pv = ((ref_skill - init_skill) > 0).mean("iteration")
    if not metric.positive:
        pv = 1 - pv
    return pv


def bootstrap_uninitialized_ensemble(hind, hist):
    """Resample uninitialized hindcast from historical members.

    Note:
        Needed for bootstrapping confidence intervals and p_values of a metric in
        the hindcast framework. Takes hind.lead.size timesteps from historical at
        same forcing and rearranges them into ensemble and member dimensions.

    Args:
        hind (xarray object): hindcast.
        hist (xarray object): historical uninitialized.

    Returns:
        uninit_hind (xarray object): uninitialize hindcast with hind.coords.
    """
    has_dims(hist, "member", "historical ensemble")
    has_dims(hind, "member", "initialized hindcast ensemble")
    # Put this after `convert_time_index` since it assigns 'years' attribute if the
    # `init` dimension is a `float` or `int`.
    has_valid_lead_units(hind)

    # find range for bootstrapping
    first_init = max(hist.time.min(), hind["init"].min())

    n, freq = get_lead_cftime_shift_args(hind.lead.attrs["units"], hind.lead.size)
    hist_last = shift_cftime_singular(hist.time.max(), -1 * n, freq)
    last_init = min(hist_last, hind["init"].max())

    hind = hind.sel(init=slice(first_init, last_init))

    uninit_hind = []
    for init in hind.init.values:
        # take uninitialized members from hist at init forcing
        # (Goddard et al. allows 5 year forcing range here)
        uninit_at_one_init_year = hist.sel(
            time=slice(
                shift_cftime_singular(init, 1, freq),
                shift_cftime_singular(init, n, freq),
            ),
        ).rename({"time": "lead"})
        uninit_at_one_init_year["lead"] = np.arange(
            1, 1 + uninit_at_one_init_year["lead"].size
        )
        uninit_hind.append(uninit_at_one_init_year)
    uninit_hind = xr.concat(uninit_hind, "init")
    uninit_hind["init"] = hind["init"].values
    uninit_hind.lead.attrs["units"] = hind.lead.attrs["units"]
    uninit_hind["member"] = hist["member"].values
    return (
        _transpose_and_rechunk_to(
            uninit_hind, hind.isel(member=[0] * uninit_hind.member.size)
        )
        if dask.is_dask_collection(uninit_hind)
        else uninit_hind
    )


def bootstrap_uninit_pm_ensemble_from_control_cftime(init_pm, control):
    """Create a pseudo-ensemble from control run.

    Bootstrap random numbers for years to construct an uninitialized ensemble from.
    This assumes a continous control simulation without gaps.

    Note:
        Needed for block bootstrapping  a metric in perfect-model framework. Takes
        random segments of length ``block_length`` from control based on ``dayofyear``
        (and therefore assumes a constant climate control simulation) and rearranges
        them into ensemble and member dimensions.

    Args:
        init_pm (xarray object): initialized ensemble simulation.
        control (xarray object): control simulation.

    Returns:
        uninit_pm (xarray object): uninitialized ensemble generated from control run.
    """
    lead_units_equal_control_time_stride(init_pm, control)
    # short cut if annual leads
    if init_pm.lead.attrs["units"] == "years":
        return _bootstrap_by_stacking(init_pm, control)

    block_length = init_pm.lead.size
    freq = get_lead_cftime_shift_args(init_pm.lead.attrs["units"], block_length)[1]
    nmember = init_pm.member.size
    # start and end years possible to resample the actual uninitialized ensembles from
    c_start_year = control.time.min().dt.year.astype("int")
    # dont resample from years that control wont have timesteps for all leads
    c_end_year = (
        shift_cftime_singular(control.time.max(), -block_length, freq).dt.year.astype(
            "int"
        )
        - 1
    )

    def sel_time(start_year_int, suitable_start_dates):
        """Select time segments from control from ``suitable_start_dates`` based on
        year ``start_year_int``."""
        start_time = suitable_start_dates.time.sel(time=str(start_year_int))
        end_time = shift_cftime_singular(start_time, block_length - 1, freq)
        new = control.sel(time=slice(*start_time, *end_time))
        new["time"] = init_pm.lead.values
        return new

    def create_pseudo_members(init):
        """For every initialization take a different set of start years."""
        startlist = np.random.randint(c_start_year, c_end_year, nmember)
        suitable_start_dates = find_start_dates_for_given_init(control, init)
        return xr.concat(
            (sel_time(start, suitable_start_dates) for start in startlist),
            dim="member",
            **CONCAT_KWARGS,
        )

    uninit = xr.concat(
        (create_pseudo_members(init) for init in init_pm.init),
        dim="init",
        **CONCAT_KWARGS,
    ).rename({"time": "lead"})
    uninit["member"] = init_pm.member.values
    uninit["lead"] = init_pm.lead
    # chunk to same dims
    transpose_kwargs = (
        {"transpose_coords": False} if isinstance(init_pm, xr.DataArray) else {}
    )
    uninit = uninit.transpose(*init_pm.dims, **transpose_kwargs)
    return (
        _transpose_and_rechunk_to(uninit, init_pm)
        if dask.is_dask_collection(uninit)
        else uninit
    )


def _bootstrap_by_stacking(init_pm, control):
    """Bootstrap member, lead, init from control by reshaping. Fast track of function
    `bootstrap_uninit_pm_ensemble_from_control_cftime` when lead units is 'years'."""
    assert type(init_pm) == type(control)
    lead_unit = init_pm.lead.attrs["units"]
    if isinstance(init_pm, xr.Dataset):
        init_pm = init_pm.to_array()
        init_was_dataset = True
    else:
        init_was_dataset = False
    if isinstance(control, xr.Dataset):
        control = control.to_array()

    init_size = init_pm.init.size * init_pm.member.size * init_pm.lead.size
    # select random start points
    new_time = np.random.randint(
        0, control.time.size - init_pm.lead.size, init_size // (init_pm.lead.size)
    )
    new_time = np.array(
        [np.arange(s, s + init_pm.lead.size) for s in new_time]
    ).flatten()[:init_size]
    larger = control.isel(time=new_time)
    fake_init = init_pm.stack(time=tuple(d for d in init_pm.dims if d in CLIMPRED_DIMS))
    # exchange values
    transpose_kwargs = (
        {"transpose_coords": False} if isinstance(init_pm, xr.DataArray) else {}
    )
    larger = larger.transpose(*fake_init.dims, **transpose_kwargs)
    fake_init.data = larger.data
    fake_uninit = fake_init.unstack()
    if init_was_dataset:
        fake_uninit = fake_uninit.to_dataset(dim="variable")
    fake_uninit["lead"] = init_pm["lead"]
    fake_uninit.lead.attrs["units"] = lead_unit
    return fake_uninit


def _bootstrap_hindcast_over_init_dim(
    hind,
    hist,
    verif,
    dim,
    reference,
    resample_dim,
    iterations,
    metric,
    comparison,
    compute,
    resample_uninit,
    **metric_kwargs,
):
    """Bootstrap hindcast skill over the ``init`` dimension.

    When bootstrapping over the ``member`` dimension, an additional dimension
    ``iteration`` can be added and skill can be computing over that entire
    dimension in parallel, since all members are being aligned the same way.
    However, to our knowledge, when bootstrapping over the ``init`` dimension,
    one must evaluate each iteration independently. I.e., in a looped fashion,
    since alignment of initializations and target dates is unique to each
    iteration.

    See ``bootstrap_compute`` for explanation of inputs.
    """
    pers_skill = []
    bootstrapped_init_skill = []
    bootstrapped_uninit_skill = []
    for i in range(iterations):
        # resample with replacement
        smp_hind = _resample(hind, resample_dim)
        # compute init skill
        init_skill = compute(
            smp_hind,
            verif,
            metric=metric,
            comparison=comparison,
            add_attrs=False,
            dim=dim,
            **metric_kwargs,
        )
        # reset inits when probabilistic, otherwise tests fail
        if (
            resample_dim == "init"
            and metric.probabilistic
            and "init" in init_skill.coords
        ):
            init_skill["init"] = hind.init.values
        bootstrapped_init_skill.append(init_skill)
        if "uninitialized" in reference:
            # generate uninitialized ensemble from hist
            uninit_hind = resample_uninit(hind, hist)
            # compute uninit skill
            bootstrapped_uninit_skill.append(
                compute(
                    uninit_hind,
                    verif,
                    metric=metric,
                    comparison=comparison,
                    dim=dim,
                    add_attrs=False,
                    **metric_kwargs,
                )
            )
        if "persistence" in reference:
            pers_skill.append(
                compute_persistence(
                    smp_hind,
                    verif,
                    metric=metric,
                    dim=dim,
                    add_attrs=False,
                    **metric_kwargs,
                )
            )
    bootstrapped_init_skill = xr.concat(
        bootstrapped_init_skill, dim="iteration", **CONCAT_KWARGS
    )
    if "uninitialized" in reference:
        bootstrapped_uninit_skill = xr.concat(
            bootstrapped_uninit_skill, dim="iteration", **CONCAT_KWARGS
        )
    else:
        bootstrapped_uninit_skill = None
    if "persistence" in reference:
        bootstrapped_pers_skill = xr.concat(
            pers_skill, dim="iteration", **CONCAT_KWARGS
        )
    else:
        bootstrapped_pers_skill = None
    return (bootstrapped_init_skill, bootstrapped_uninit_skill, bootstrapped_pers_skill)


def _get_resample_func(ds):
    """Decide for resample function based on input `ds`.

    Returns:
      callable: `_resample_iterations`: if big and chunked `ds`
                `_resample_iterations_idx`: else (if small and eager `ds`)
    """
    resample_func = (
        _resample_iterations
        if (
            dask.is_dask_collection(ds)
            and len(ds.dims) > 3
            # > 2MB
            and ds.nbytes > 2000000
        )
        else _resample_iterations_idx
    )
    return resample_func


def _maybe_auto_chunk(ds, dims):
    """Auto-chunk on dimension `dims`.

    Args:
        ds (xr.object): input data.
        dims (list of str or str): Dimensions to auto-chunk in.

    Returns:
        xr.object: auto-chunked along `dims`

    """
    if dask.is_dask_collection(ds) and dims is not []:
        if isinstance(dims, str):
            dims = [dims]
        chunks = [d for d in dims if d in ds.dims]
        chunks = {key: "auto" for key in chunks}
        ds = ds.chunk(chunks)
    return ds


def _chunk_before_resample_iterations_idx(
    ds, iterations, chunking_dims, optimal_blocksize=100000000
):
    """Chunk ds so small that after _resample_iteration_idx chunks have optimal size
    `optimal_blocksize`.

    Args:
        ds (xr.obejct): input data`.
        iterations (int): number of bootstrap iterations in `_resample_iterations_idx`.
        chunking_dims (list of str or str): Dimension(s) to chunking in.
        optimal_blocksize (int): dask blocksize to aim at in bytes.
            Defaults to 100000000.

    Returns:
        xr.object: chunked to have blocksize: optimal_blocksize/iterations.

    """
    if isinstance(chunking_dims, str):
        chunking_dims = [chunking_dims]
    # size of CLIMPRED_DIMS
    climpred_dim_chunksize = 8 * np.product(
        np.array([ds[d].size for d in CLIMPRED_DIMS if d in ds.dims])
    )
    # remaining blocksize for remaining dims considering iteration
    spatial_dim_blocksize = optimal_blocksize / (climpred_dim_chunksize * iterations)
    # size of remaining dims
    chunking_dims_size = np.product(
        np.array([ds[d].size for d in ds.dims if d not in CLIMPRED_DIMS])
    )  # ds.lat.size*ds.lon.size
    # chunks needed to get to optimal blocksize
    chunks_needed = chunking_dims_size / spatial_dim_blocksize
    # get size clon, clat for spatial chunks
    cdim = [1 for i in chunking_dims]
    nchunks = np.product(cdim)
    stepsize = 1
    counter = 0
    while nchunks < chunks_needed:
        for i, d in enumerate(chunking_dims):
            c = cdim[i]
            if c <= ds[d].size:
                c = c + stepsize
                cdim[i] = c
            nchunks = np.product(cdim)
        counter += 1
        if counter == 100:
            break
    # convert number of chunks to chunksize
    chunks = dict()
    for i, d in enumerate(chunking_dims):
        chunksize = ds[d].size // cdim[i]
        if chunksize < 1:
            chunksize = 1
        chunks[d] = chunksize
    ds = ds.chunk(chunks)
    return ds


def bootstrap_compute(
    hind,
    verif,
    hist=None,
    alignment="same_verifs",
    metric="pearson_r",
    comparison="m2e",
    dim="init",
    reference=None,
    resample_dim="member",
    sig=95,
    iterations=500,
    pers_sig=None,
    compute=compute_hindcast,
    resample_uninit=bootstrap_uninitialized_ensemble,
    **metric_kwargs,
):
    """Bootstrap compute with replacement.

    Args:
        hind (xr.Dataset): prediction ensemble.
        verif (xr.Dataset): Verification data.
        hist (xr.Dataset): historical/uninitialized simulation.
        metric (str): `metric`. Defaults to 'pearson_r'.
        comparison (str): `comparison`. Defaults to 'm2e'.
        dim (str or list): dimension(s) to apply metric over. default: 'init'.
        reference (str, list of str): Type of reference forecasts with which to
            verify. One or more of ['persistence', 'uninitialized'].
            If None or empty, returns no p value.
        resample_dim (str): dimension to resample from. default: 'member'::

            - 'member': select a different set of members from hind
            - 'init': select a different set of initializations from hind

        sig (int): Significance level for uninitialized and
                   initialized skill. Defaults to 95.
        pers_sig (int): Significance level for persistence skill confidence levels.
                        Defaults to sig.
        iterations (int): number of resampling iterations (bootstrap
                         with replacement). Defaults to 500.
        compute (func): function to compute skill.
                        Choose from
                        [:py:func:`climpred.prediction.compute_perfect_model`,
                         :py:func:`climpred.prediction.compute_hindcast`].
        resample_uninit (func): function to create an uninitialized ensemble
                        from a control simulation or uninitialized large
                        ensemble. Choose from:
                        [:py:func:`bootstrap_uninitialized_ensemble`,
                         :py:func:`bootstrap_uninit_pm_ensemble_from_control`].
        ** metric_kwargs (dict): additional keywords to be passed to metric
            (see the arguments required for a given metric in :ref:`Metrics`).

    Returns:
        results: (xr.Dataset): bootstrapped results for the three different skills:

            - `initialized` for the initialized hindcast `hind` and describes skill due
             to initialization and external forcing
            - `uninitialized` for the uninitialized/historical and approximates skill
             from external forcing
            - `persistence` for the persistence forecast computed by
              `compute_persistence`

        the different results:
            - `verify skill`: skill values
            - `p`: p value
            - `low_ci` and `high_ci`: high and low ends of confidence intervals based
             on significance threshold `sig`


    Reference:
        * Goddard, L., A. Kumar, A. Solomon, D. Smith, G. Boer, P.
          Gonzalez, V. Kharin, et al. “A Verification Framework for
          Interannual-to-Decadal Predictions Experiments.” Climate
          Dynamics 40, no. 1–2 (January 1, 2013): 245–72.
          https://doi.org/10/f4jjvf.

    See also:
        * climpred.bootstrap.bootstrap_hindcast
        * climpred.bootstrap.bootstrap_perfect_model
    """
    warn_if_chunking_would_increase_performance(hind, crit_size_in_MB=5)
    if pers_sig is None:
        pers_sig = sig
    if isinstance(dim, str):
        dim = [dim]
    if isinstance(reference, str):
        reference = [reference]
    if reference is None:
        reference = []

    p = (100 - sig) / 100
    ci_low = p / 2
    ci_high = 1 - p / 2
    p_pers = (100 - pers_sig) / 100
    ci_low_pers = p_pers / 2
    ci_high_pers = 1 - p_pers / 2

    # get metric/comparison function name, not the alias
    metric = METRIC_ALIASES.get(metric, metric)
    comparison = COMPARISON_ALIASES.get(comparison, comparison)

    # get class Metric(metric)
    metric = get_metric_class(metric, ALL_METRICS)
    # get comparison function
    comparison = get_comparison_class(comparison, ALL_COMPARISONS)

    # Perfect Model requires `same_inits` setup
    isHindcast = True if comparison.name in HINDCAST_COMPARISONS else False
    reference_alignment = alignment if isHindcast else "same_inits"
    chunking_dims = [d for d in hind.dims if d not in CLIMPRED_DIMS]

    # carry alignment for compute_reference separately
    metric_kwargs_reference = metric_kwargs.copy()
    metric_kwargs_reference["alignment"] = reference_alignment
    # carry alignment in metric_kwargs
    if isHindcast:
        metric_kwargs["alignment"] = alignment

    if hist is None:  # PM path, use verif = control
        hist = verif

    # slower path for hindcast and resample_dim init
    if resample_dim == "init" and isHindcast:
        warnings.warn("resample_dim=`init` will be slower than resample_dim=`member`.")
        (
            bootstrapped_init_skill,
            bootstrapped_uninit_skill,
            bootstrapped_pers_skill,
        ) = _bootstrap_hindcast_over_init_dim(
            hind,
            hist,
            verif,
            dim,
            reference,
            resample_dim,
            iterations,
            metric,
            comparison,
            compute,
            resample_uninit,
            **metric_kwargs,
        )
    else:  # faster: first _resample_iterations_idx, then compute skill
        resample_func = _get_resample_func(hind)
        if not isHindcast:
            if "uninitialized" in reference:
                # create more members than needed in PM to make the uninitialized
                # distribution more robust
                members_to_sample_from = 50
                repeat = members_to_sample_from // hind.member.size + 1
                uninit_hind = xr.concat(
                    [resample_uninit(hind, hist) for i in range(repeat)],
                    dim="member",
                    **CONCAT_KWARGS,
                )
                uninit_hind["member"] = np.arange(1, 1 + uninit_hind.member.size)
                if dask.is_dask_collection(uninit_hind):
                    # too minimize tasks: ensure uninit_hind get pre-computed
                    # alternativly .chunk({'member':-1})
                    uninit_hind = uninit_hind.compute().chunk()
                # resample uninit always over member and select only hind.member.size
                bootstrapped_uninit = resample_func(
                    uninit_hind,
                    iterations,
                    "member",
                    replace=False,
                    dim_max=hind["member"].size,
                )
                bootstrapped_uninit["lead"] = hind["lead"]
                # effectively only when _resample_iteration_idx which doesnt use dim_max
                bootstrapped_uninit = bootstrapped_uninit.isel(
                    member=slice(None, hind.member.size)
                )
                bootstrapped_uninit["member"] = np.arange(
                    1, 1 + bootstrapped_uninit.member.size
                )
                if dask.is_dask_collection(bootstrapped_uninit):
                    bootstrapped_uninit = bootstrapped_uninit.chunk({"member": -1})
                    bootstrapped_uninit = _maybe_auto_chunk(
                        bootstrapped_uninit, ["iteration"] + chunking_dims
                    )
        else:  # hindcast
            if "uninitialized" in reference:
                uninit_hind = resample_uninit(hind, hist)
                if dask.is_dask_collection(uninit_hind):
                    # too minimize tasks: ensure uninit_hind get pre-computed
                    # maybe not needed
                    uninit_hind = uninit_hind.compute().chunk()
                bootstrapped_uninit = resample_func(
                    uninit_hind, iterations, resample_dim
                )
                bootstrapped_uninit = bootstrapped_uninit.isel(
                    member=slice(None, hind.member.size)
                )
                bootstrapped_uninit["lead"] = hind["lead"]
                if dask.is_dask_collection(bootstrapped_uninit):
                    bootstrapped_uninit = _maybe_auto_chunk(
                        bootstrapped_uninit.chunk({"lead": 1}),
                        ["iteration"] + chunking_dims,
                    )

        if "uninitialized" in reference:
            bootstrapped_uninit_skill = compute(
                bootstrapped_uninit,
                verif,
                metric=metric,
                comparison="m2o" if isHindcast else comparison,
                dim=dim,
                add_attrs=False,
                **metric_kwargs,
            )
            # take mean if 'm2o' comparison forced before
            if isHindcast and comparison != __m2o:
                bootstrapped_uninit_skill = bootstrapped_uninit_skill.mean("member")

        with xr.set_options(keep_attrs=True):
            bootstrapped_hind = resample_func(hind, iterations, resample_dim)
        if dask.is_dask_collection(bootstrapped_hind):
            bootstrapped_hind = bootstrapped_hind.chunk({"member": -1})

        bootstrapped_init_skill = compute(
            bootstrapped_hind,
            verif,
            metric=metric,
            comparison=comparison,
            add_attrs=False,
            dim=dim,
            **metric_kwargs,
        )
        if "persistence" in reference:
            pers_skill = compute_persistence(
                hind,
                verif,
                metric=metric,
                dim=dim,
                **metric_kwargs_reference,
            )
            # bootstrap pers
            if resample_dim == "init":
                bootstrapped_pers_skill = compute_persistence(
                    bootstrapped_hind,
                    verif,
                    metric=metric,
                    **metric_kwargs_reference,
                )
            else:  # member no need to calculate all again
                bootstrapped_pers_skill, _ = xr.broadcast(
                    pers_skill, bootstrapped_init_skill
                )

    # calc mean skill without any resampling
    init_skill = compute(
        hind,
        verif,
        metric=metric,
        comparison=comparison,
        dim=dim,
        **metric_kwargs,
    )
    if "uninitialized" in reference:
        # uninit skill as mean resampled uninit skill
        unin_skill = bootstrapped_uninit_skill.mean("iteration")  # noqa: F841
    if "persistence" in reference:
        pers_skill = compute_persistence(
            hind, verif, metric=metric, dim=dim, **metric_kwargs_reference
        )
    if "climatology" in reference:
        clim_skill = compute_climatology(
            hind, verif, metric=metric, dim=dim, comparison=comparison, **metric_kwargs
        )
        bootstrapped_clim_skill, _ = xr.broadcast(clim_skill, bootstrapped_init_skill)

    # get confidence intervals CI
    init_ci = _distribution_to_ci(bootstrapped_init_skill, ci_low, ci_high)
    if "uninitialized" in reference:
        unin_ci = _distribution_to_ci(  # noqa: F841
            bootstrapped_uninit_skill, ci_low, ci_high
        )
    if "climatology" in reference:
        clim_ci = _distribution_to_ci(  # noqa: F841
            bootstrapped_clim_skill, ci_low, ci_high
        )
    if "persistence" in reference:
        pers_ci = _distribution_to_ci(  # noqa: F841
            bootstrapped_pers_skill, ci_low_pers, ci_high_pers
        )

    # pvalue whether uninit or pers better than init forecast
    if "uninitialized" in reference:
        p_unin_over_init = _pvalue_from_distributions(  # noqa: F841
            bootstrapped_uninit_skill, bootstrapped_init_skill, metric=metric
        )
    if "climatology" in reference:
        p_clim_over_init = _pvalue_from_distributions(  # noqa: F841
            bootstrapped_clim_skill, bootstrapped_init_skill, metric=metric
        )
    if "persistence" in reference:
        p_pers_over_init = _pvalue_from_distributions(  # noqa: F841
            bootstrapped_pers_skill, bootstrapped_init_skill, metric=metric
        )

    # gather return
    # p defined as probability that reference better than
    # initialized, therefore not defined for initialized skill
    # itself
    results = xr.concat(
        [
            init_skill,
            init_skill.where(init_skill == -999),
            init_ci.isel(quantile=0, drop=True),
            init_ci.isel(quantile=1, drop=True),
        ],
        dim="results",
        coords="minimal",
    ).assign_coords(
        results=("results", ["verify skill", "p", "low_ci", "high_ci"]),
        skill="initialized",
    )

    if reference != []:
        for r in reference:
            ref_skill = eval(f"{r[:4]}_skill")
            ref_p = eval(f"p_{r[:4]}_over_init")
            ref_ci_low = eval(f"{r[:4]}_ci").isel(quantile=0, drop=True)
            ref_ci_high = eval(f"{r[:4]}_ci").isel(quantile=1, drop=True)
            ref_results = xr.concat(
                [ref_skill, ref_p, ref_ci_low, ref_ci_high],
                dim="results",
                **CONCAT_KWARGS,
            ).assign_coords(
                skill=r, results=("results", ["verify skill", "p", "low_ci", "high_ci"])
            )
            if "member" in ref_results.dims:
                if not ref_results["member"].identical(results["member"]):
                    ref_results["member"] = results[
                        "member"
                    ]  # fixes m2c different member names in reference forecasts
            results = xr.concat([results, ref_results], dim="skill", **CONCAT_KWARGS)
        results = results.assign_coords(skill=["initialized"] + reference).squeeze()
    else:
        results = results.drop_sel(results="p")
    results = results.squeeze()

    # Attach climpred compute information to skill
    # results.results
    metadata_dict = {
        "confidence_interval_levels": f"{ci_high}-{ci_low}",
        "bootstrap_iterations": iterations,
    }
    if reference is not None:
        metadata_dict[
            "p"
        ] = "probability that reference performs better than initialized"
    metadata_dict.update(metric_kwargs)
    results = assign_attrs(
        results,
        hind,
        alignment=alignment,
        metric=metric,
        comparison=comparison,
        dim=dim,
        metadata_dict=metadata_dict,
    )
    # Ensure that the lead units get carried along for the calculation. The attribute
    # tends to get dropped along the way due to ``xarray`` functionality.
    results["lead"] = hind["lead"]
    if "units" in hind["lead"].attrs and "units" not in results["lead"].attrs:
        results["lead"].attrs["units"] = hind["lead"].attrs["units"]
    return results


def bootstrap_hindcast(
    hind,
    hist,
    verif,
    alignment="same_verifs",
    metric="pearson_r",
    comparison="e2o",
    dim="init",
    reference=None,
    resample_dim="member",
    sig=95,
    iterations=500,
    pers_sig=None,
    **metric_kwargs,
):
    """Bootstrap compute with replacement. Wrapper of
     py:func:`bootstrap_compute` for hindcasts.

    Args:
        hind (xr.Dataset): prediction ensemble.
        verif (xr.Dataset): Verification data.
        hist (xr.Dataset): historical/uninitialized simulation.
        metric (str): `metric`. Defaults to 'pearson_r'.
        comparison (str): `comparison`. Defaults to 'e2o'.
        dim (str): dimension to apply metric over. default: 'init'.
        reference (str, list of str): Type of reference forecasts with which to
            verify. One or more of ['persistence', 'uninitialized'].
            If None or empty, returns no p value.
        resample_dim (str or list): dimension to resample from. default: 'member'.

            - 'member': select a different set of members from hind
            - 'init': select a different set of initializations from hind

        sig (int): Significance level for uninitialized and
                   initialized skill. Defaults to 95.
        pers_sig (int): Significance level for persistence skill confidence levels.
                        Defaults to sig.
        iterations (int): number of resampling iterations (bootstrap
                         with replacement). Defaults to 500.
        ** metric_kwargs (dict): additional keywords to be passed to metric
            (see the arguments required for a given metric in :ref:`Metrics`).

    Returns:
        results: (xr.Dataset): bootstrapped results for the three different kinds of
                               predictions:

            - `initialized` for the initialized hindcast `hind` and describes skill due
             to initialization and external forcing
            - `uninitialized` for the uninitialized/historical and approximates skill
             from external forcing
            - `persistence` for the persistence forecast computed by
             `compute_persistence`

        the different results:
            - `verify skill`: skill values
            - `p`: p value
            - `low_ci` and `high_ci`: high and low ends of confidence intervals based
             on significance threshold `sig`

    Reference:
        * Goddard, L., A. Kumar, A. Solomon, D. Smith, G. Boer, P.
          Gonzalez, V. Kharin, et al. “A Verification Framework for
          Interannual-to-Decadal Predictions Experiments.” Climate
          Dynamics 40, no. 1–2 (January 1, 2013): 245–72.
          https://doi.org/10/f4jjvf.

    See also:
        * climpred.bootstrap.bootstrap_compute
        * climpred.prediction.compute_hindcast

    """
    # Check that init is int, cftime, or datetime; convert ints or datetime to cftime.
    hind = convert_time_index(hind, "init", "hind[init]")
    if isinstance(hist, xr.Dataset):
        hist = convert_time_index(hist, "time", "uninitialized[time]")
    else:
        hist = False
    verif = convert_time_index(verif, "time", "verif[time]")
    # Put this after `convert_time_index` since it assigns 'years' attribute if the
    # `init` dimension is a `float` or `int`.
    has_valid_lead_units(hind)

    if ("same_verif" in alignment) & (resample_dim == "init"):
        raise KeywordError(
            "Cannot have both alignment='same_verifs' and "
            "resample_dim='init'. Change `resample_dim` to 'member' to keep "
            "common verification alignment or `alignment` to 'same_inits' to "
            "resample over initializations."
        )

    # Kludge for now. Since we're computing persistence here we need to ensure that
    # all products have a union in their time axis.
    if hist not in [None, False]:
        times = np.sort(
            list(set(hind.init.data) & set(hist.time.data) & set(verif.time.data))
        )
    else:
        times = np.sort(list(set(hind.init.data) & set(verif.time.data)))
    hind = hind.sel(init=times)
    if isinstance(hist, xr.Dataset):
        hist = hist.sel(time=times)
    verif = verif.sel(time=times)

    return bootstrap_compute(
        hind,
        verif,
        hist=hist,
        alignment=alignment,
        metric=metric,
        comparison=comparison,
        dim=dim,
        reference=reference,
        resample_dim=resample_dim,
        sig=sig,
        iterations=iterations,
        pers_sig=pers_sig,
        compute=compute_hindcast,
        resample_uninit=bootstrap_uninitialized_ensemble,
        **metric_kwargs,
    )


def bootstrap_perfect_model(
    init_pm,
    control,
    metric="pearson_r",
    comparison="m2e",
    dim=None,
    reference=None,
    resample_dim="member",
    sig=95,
    iterations=500,
    pers_sig=None,
    **metric_kwargs,
):
    """Bootstrap compute with replacement. Wrapper of
     py:func:`bootstrap_compute` for perfect-model framework.

    Args:
        hind (xr.Dataset): prediction ensemble.
        verif (xr.Dataset): Verification data.
        hist (xr.Dataset): historical/uninitialized simulation.
        metric (str): `metric`. Defaults to 'pearson_r'.
        comparison (str): `comparison`. Defaults to 'm2e'.
        dim (str): dimension to apply metric over. default: ['init', 'member'].
        reference (str, list of str): Type of reference forecasts with which to
            verify. One or more of ['persistence', 'uninitialized'].
            If None or empty, returns no p value.
        resample_dim (str or list): dimension to resample from. default: 'member'.

            - 'member': select a different set of members from hind
            - 'init': select a different set of initializations from hind

        sig (int): Significance level for uninitialized and
                   initialized skill. Defaults to 95.
        pers_sig (int): Significance level for persistence skill confidence levels.
                        Defaults to sig.
        iterations (int): number of resampling iterations (bootstrap
                         with replacement). Defaults to 500.
        ** metric_kwargs (dict): additional keywords to be passed to metric
            (see the arguments required for a given metric in :ref:`Metrics`).

    Returns:
        results: (xr.Dataset): bootstrapped results for the three different kinds of
                               predictions:

            - `initialized` for the initialized hindcast `hind` and describes skill due
             to initialization and external forcing
            - `uninitialized` for the uninitialized/historical and approximates skill
             from external forcing
            - `persistence` for the persistence forecast
            - `climatology`

        the different results:
            - `skill`: skill values
            - `p`: p value
            - `low_ci` and `high_ci`: high and low ends of confidence intervals based
             on significance threshold `sig`

    Reference:
        * Goddard, L., A. Kumar, A. Solomon, D. Smith, G. Boer, P.
          Gonzalez, V. Kharin, et al. “A Verification Framework for
          Interannual-to-Decadal Predictions Experiments.” Climate
          Dynamics 40, no. 1–2 (January 1, 2013): 245–72.
          https://doi.org/10/f4jjvf.

    See also:
        * climpred.bootstrap.bootstrap_compute
        * climpred.prediction.compute_perfect_model
    """
    if dim is None:
        dim = ["init", "member"]
    # Check init & time is int, cftime, or datetime; convert ints or datetime to cftime.
    init_pm = convert_time_index(
        init_pm, "init", "init_pm[init]", calendar=PM_CALENDAR_STR
    )
    control = convert_time_index(
        control, "time", "control[time]", calendar=PM_CALENDAR_STR
    )
    lead_units_equal_control_time_stride(init_pm, control)
    return bootstrap_compute(
        init_pm,
        control,
        hist=None,
        metric=metric,
        comparison=comparison,
        dim=dim,
        reference=reference,
        resample_dim=resample_dim,
        sig=sig,
        iterations=iterations,
        pers_sig=pers_sig,
        compute=compute_perfect_model,
        resample_uninit=bootstrap_uninit_pm_ensemble_from_control_cftime,
        **metric_kwargs,
    )


def _bootstrap_func(
    func,
    ds,
    resample_dim,
    sig=95,
    iterations=500,
    *func_args,
    **func_kwargs,
):
    """Sig % threshold of function based on iterations resampling with replacement.

    Reference:
    * Mason, S. J., and G. M. Mimmack. “The Use of Bootstrap Confidence
     Intervals for the Correlation Coefficient in Climatology.” Theoretical and
      Applied Climatology 45, no. 4 (December 1, 1992): 229–33.
      https://doi.org/10/b6fnsv.

    Args:
        func (function): function to be bootstrapped.
        ds (xr.object): first input argument of func. `chunk` ds on `dim` other
            than `resample_dim` for potential performance increase when multiple
            CPUs available.
        resample_dim (str): dimension to resample from.
        sig (int,float,list): significance levels to return. Defaults to 95.
        iterations (int): number of resample iterations. Defaults to 500.
        *func_args (type): `*func_args`.
        **func_kwargs (type): `**func_kwargs`.

    Returns:
        sig_level: bootstrapped significance levels with
                   dimensions of init_pm and len(sig) if sig is list
    """
    if not callable(func):
        raise ValueError(f"Please provide func as a function, found {type(func)}")
    warn_if_chunking_would_increase_performance(ds)
    if isinstance(sig, list):
        psig = [i / 100 for i in sig]
    else:
        psig = sig / 100

    resample_func = _get_resample_func(ds)
    bootstraped_ds = resample_func(ds, iterations, dim=resample_dim, replace=False)
    bootstraped_results = func(bootstraped_ds, *func_args, **func_kwargs)
    bootstraped_results = rechunk_to_single_chunk_if_more_than_one_chunk_along_dim(
        bootstraped_results, dim="iteration"
    )
    sig_level = bootstraped_results.quantile(dim="iteration", q=psig, skipna=False)
    return sig_level


def dpp_threshold(control, sig=95, iterations=500, dim="time", **dpp_kwargs):
    """Calc DPP significance levels from re-sampled dataset.

    Reference:
        * Feng, X., T. DelSole, and P. Houser. “Bootstrap Estimated Seasonal
          Potential Predictability of Global Temperature and Precipitation.”
          Geophysical Research Letters 38, no. 7 (2011).
          https://doi.org/10/ft272w.

    See also:
        * climpred.bootstrap._bootstrap_func
        * climpred.stats.dpp
    """
    return _bootstrap_func(
        dpp, control, dim, sig=sig, iterations=iterations, **dpp_kwargs
    )


def varweighted_mean_period_threshold(control, sig=95, iterations=500, time_dim="time"):
    """Calc the variance-weighted mean period significance levels from re-sampled
    dataset.

    See also:
        * climpred.bootstrap._bootstrap_func
        * climpred.stats.varweighted_mean_period
    """
    if varweighted_mean_period is None:
        raise ImportError(
            "xrft is not installed; see"
            "https://xrft.readthedocs.io/en/latest/installation.html"
        )
    return _bootstrap_func(
        varweighted_mean_period,
        control,
        time_dim,
        sig=sig,
        iterations=iterations,
    )



import numpy as np
import xarray as xr
from dask.distributed import Client

from climpred.bootstrap import bootstrap_perfect_model
from climpred.metrics import PROBABILISTIC_METRICS
from climpred.prediction import compute_perfect_model

from . import ensure_loaded, parameterized, randn, requires_dask

# only take subselection of all possible metrics
METRICS = ["rmse", "pearson_r", "crpss"]
# only take comparisons compatible with probabilistic metrics
PM_COMPARISONS = ["m2m", "m2c"]

ITERATIONS = 16


class Generate:
    """
    Generate input data for benchmark."""

    timeout = 600
    repeat = (2, 5, 20)

    def make_initialized_control(self):
        """Generates initialized ensembles and a control simulation, mimicking a
        perfect-model experiment."""
        self.ds = xr.Dataset()
        self.control = xr.Dataset()
        self.nmember = 5
        self.ninit = 6
        self.nlead = 10
        self.iterations = ITERATIONS
        self.nx = 72
        self.ny = 36
        self.control_start = 3000
        self.control_end = 3300
        self.ntime = self.control_end - self.control_start
        self.client = None

        FRAC_NAN = 0.0

        times = xr.cftime_range(
            start=str(self.control_start),
            periods=self.ntime,
            freq="YS",
            calendar="noleap",
        )
        leads = np.arange(1, 1 + self.nlead)
        members = np.arange(1, 1 + self.nmember)
        inits = xr.cftime_range(
            start=str(self.control_start),
            periods=self.ninit,
            freq="10YS",
            calendar="noleap",
        )

        lons = xr.DataArray(
            np.linspace(0.5, 359.5, self.nx),
            dims=("lon",),
            attrs={"units": "degrees east", "long_name": "longitude"},
        )
        lats = xr.DataArray(
            np.linspace(-89.5, 89.5, self.ny),
            dims=("lat",),
            attrs={"units": "degrees north", "long_name": "latitude"},
        )
        self.ds["var"] = xr.DataArray(
            randn(
                (self.nmember, self.ninit, self.nlead, self.nx, self.ny),
                frac_nan=FRAC_NAN,
            ),
            coords={
                "member": members,
                "init": inits,
                "lon": lons,
                "lat": lats,
                "lead": leads,
            },
            dims=("member", "init", "lead", "lon", "lat"),
            name="var",
            attrs={"units": "var units", "description": "a description"},
        )
        self.control["var"] = xr.DataArray(
            randn((self.ntime, self.nx, self.ny), frac_nan=FRAC_NAN),
            coords={"lon": lons, "lat": lats, "time": times},
            dims=("time", "lon", "lat"),
            name="var",
            attrs={"units": "var units", "description": "a description"},
        )

        self.ds.attrs = {"history": "created for xarray benchmarking"}
        self.ds.lead.attrs["units"] = "years"
        self.control.time.attrs["units"] = "years"


class Compute(Generate):
    """
    Benchmark time and peak memory of `compute_perfect_model` and
    `bootstrap_perfect_model`.
    """

    def setup(self, *args, **kwargs):
        self.make_initialized_control()

    @parameterized(["metric", "comparison"], (METRICS, PM_COMPARISONS))
    def time_compute_perfect_model(self, metric, comparison):
        """Take time for `compute_perfect_model`."""
        dim = "member" if metric in PROBABILISTIC_METRICS else None
        ensure_loaded(
            compute_perfect_model(
                self.ds, self.control, metric=metric, comparison=comparison, dim=dim
            )
        )

    @parameterized(["metric", "comparison"], (METRICS, PM_COMPARISONS))
    def peakmem_compute_perfect_model(self, metric, comparison):
        """Take memory peak for `compute_perfect_model`."""
        dim = "member" if metric in PROBABILISTIC_METRICS else None
        ensure_loaded(
            compute_perfect_model(
                self.ds, self.control, metric=metric, comparison=comparison, dim=dim
            )
        )

    @parameterized(["metric", "comparison"], (METRICS, PM_COMPARISONS))
    def time_bootstrap_perfect_model(self, metric, comparison):
        """Take time for `bootstrap_perfect_model`."""
        dim = "member" if metric in PROBABILISTIC_METRICS else None
        ensure_loaded(
            bootstrap_perfect_model(
                self.ds,
                self.control,
                metric=metric,
                comparison=comparison,
                iterations=self.iterations,
                dim=dim,
            )
        )

    @parameterized(["metric", "comparison"], (METRICS, PM_COMPARISONS))
    def peakmem_bootstrap_perfect_model(self, metric, comparison):
        """Take memory peak for `bootstrap_perfect_model`."""
        dim = "member" if metric in PROBABILISTIC_METRICS else None
        ensure_loaded(
            bootstrap_perfect_model(
                self.ds,
                self.control,
                metric=metric,
                comparison=comparison,
                iterations=self.iterations,
                dim=dim,
            )
        )


class ComputeDask(Compute):
    def setup(self, *args, **kwargs):
        """Benchmark time and peak memory of `compute_perfect_model` and
        `bootstrap_perfect_model`. This executes the same tests as `Compute` but
        on chunked data."""
        requires_dask()
        # magic taken from
        # https://github.com/pydata/xarray/blob/stable/asv_bench/benchmarks/rolling.py
        super().setup(**kwargs)
        # chunk along a spatial dimension to enable embarrasingly parallel computation
        self.ds = self.ds["var"].chunk()
        self.control = self.control["var"].chunk()


class ComputeDaskDistributed(ComputeDask):
    def setup(self, *args, **kwargs):
        """Benchmark time and peak memory of `compute_perfect_model` and
        `bootstrap_perfect_model`. This executes the same tests as `Compute` but
        on chunked data with dask.distributed.Client."""
        requires_dask()
        # magic taken from
        # https://github.com/pydata/xarray/blob/stable/asv_bench/benchmarks/rolling.py
        super().setup(**kwargs)
        self.client = Client()

    def cleanup(self):
        self.client.shutdown()


class ComputeSmall(Compute):
    def setup(self, *args, **kwargs):
        """Benchmark time and peak memory of `compute_perfect_model` and
        `bootstrap_perfect_model`. This executes the same tests as `Compute`
        but on 1D data."""
        # magic taken from
        # https://github.com/pydata/xarray/blob/stable/asv_bench/benchmarks/rolling.py
        super().setup(**kwargs)
        spatial_dims = ["lon", "lat"]
        self.ds = self.ds.mean(spatial_dims)
        self.control = self.control.mean(spatial_dims)
        self.iterations = 500



import numpy as np
import xarray as xr

from ..constants import CLIMPRED_ENSEMBLE_DIMS
from .mpi import get_path as get_path_mpi


def load_hindcast(
    inits=range(1961, 1965),
    members=range(1, 3),
    preprocess=None,
    lead_offset=1,
    parallel=True,
    engine=None,
    get_path=get_path_mpi,
    **get_path_kwargs,
):  # “pragma: no cover”
    """Load multi-member, multi-initialization hindcast experiment into one
    `xr.Dataset` compatible with `climpred`.

    Args:
        inits (list, array): List of initializations to be loaded.
            Defaults to range(1961, 1965).
        members (list, array): List of initializations to be loaded.
            Defaults to range(1, 3).
        preprocess (function): `preprocess` function accepting and returning
            `xr.Dataset` only. To be passed to :py:func:`xarray.open_dataset`.
            Defaults to None.
        parallel (bool): passed to `xr.open_mfdataset`. Defaults to True.
        engine (str): passed to `xr.open_mfdataset`. Defaults to None.

        .. note::
            To load MPI-ESM grb files, pass `engine='pynio'`.

        get_path (callable): `get_path` function specific to modelling center output
            format. Defaults to :py:func:`~climpred.preprocessing.mpi.get_path`.
        **get_path_kwargs (dict): parameters passed to `**get_path`.

    Returns:
        xr.Dataset: `climpred` compatible dataset with dims: `member`, `init`, `lead`.

    """
    init_list = []
    for init in inits:
        print(f"Processing init {init} ...")
        member_list = []
        for member in members:
            # get path p
            p = get_path(member=member, init=init, **get_path_kwargs)
            # open all leads for specified member and init
            member_ds = xr.open_mfdataset(
                p,
                combine="nested",
                concat_dim="time",
                preprocess=preprocess,
                parallel=parallel,
                engine=engine,
                coords="minimal",  # expecting identical coords
                data_vars="minimal",  # expecting identical vars
                compat="override",  # speed up
            ).squeeze()
            # set new integer time
            member_ds = set_integer_time_axis(member_ds)
            member_list.append(member_ds)
        member_ds = xr.concat(member_list, "member")
        init_list.append(member_ds)
    ds = xr.concat(init_list, "init").rename({"time": "lead"})
    ds["member"] = members
    ds["init"] = inits
    return ds


def rename_SLM_to_climpred_dims(xro):
    """Rename ensemble dimensions common to SubX or CESM output:

        * ``S`` : Refers to start date and is changed to ``init``
        * ``L`` : Refers to lead time and is changed to ``lead``
        * ``M``: Refers to ensemble member and is changed to ``member``

    Args:
        xro (xr.object): input from CESM/SubX containing dimensions: `S`, `L`, `M`.
    Returns:
        xr.object: `climpred` compatible with dimensions: `member`, `init`, `lead`.
    """
    dim_dict = {"S": "init", "L": "lead", "M": "member"}
    for dim in dim_dict.keys():
        if dim in xro.dims:
            xro = xro.rename({dim: dim_dict[dim]})
    return xro


def rename_to_climpred_dims(xro):
    """Rename existing dimension in xr.object `xro` to `CLIMPRED_ENSEMBLE_DIMS` from
    existing dimension names. This function attempts to autocorrect dimension names to
    climpred standards. e.g., `ensemble_member` becomes `member` and `lead_time`
    becomes `lead`, and `time` gets renamed to `lead`.

    Args:
        xro (xr.object): input from DCPP via `intake-esm <intake-esm.readthedocs.io/>`_
        containing dimension names like `dcpp_init_year`, `time`, `member_id`.
    Returns:
        xr.object: `climpred` compatible with dimensions: `member`, `init`, `lead`.
    """
    for cdim in CLIMPRED_ENSEMBLE_DIMS:
        renamed = False  # set renamed flag to false initiallly
        if cdim not in xro.dims:  # if a CLIMPRED_ENSEMBLE_DIMS is not found
            for c in xro.dims:  # check in xro.dims for dims
                if cdim in c:  # containing the string of this CLIMPRED_ENSEMBLE_DIMS
                    xro = xro.rename({c: cdim})
                    renamed = True
        # special case for hindcast when containing time
        if "time" in xro.dims and "lead" not in xro.dims:
            xro = xro.rename({"time": "lead"})
            renamed = True
        elif "lead" in xro.dims:
            renamed = True
        if not renamed:
            raise ValueError(
                f"Couldn't find a dimension to rename to `{cdim}`, found {xro.dims}."
            )
    return xro


def set_integer_time_axis(xro, offset=1, time_dim="time"):
    """Set time axis to integers starting from `offset`. Used in hindcast preprocessing
    before the concatination of `intake-esm` happens."""
    xro[time_dim] = np.arange(offset, offset + xro[time_dim].size)
    return xro



import numpy as np
import pytest
import xarray as xr

from climpred import HindcastEnsemble, PerfectModelEnsemble

NLEADS = 5
NMEMBERS = 3
NINITS = 10
START = "1990"

HindcastEnsemble_verify_kw = dict(
    metric="rmse", comparison="e2o", dim="init", alignment="maximize"
)
PerfectModelEnsemble_verify_kw = dict(metric="rmse", comparison="e2c", dim="init")


@pytest.fixture(
    params=[
        "seconds",
        "minutes",
        "hours",
        "days",
        "pentads",
        "weeks",
        "months",
        "seasons",
        "years",
    ]
)
def HindcastEnsemble_time_resolution(request):
    """Create HindcastEnsemble of given lead time resolution."""
    if request.param == "pentads":
        freq = "5D"
    elif request.param == "weeks":
        freq = "7D"
    elif request.param == "minutes":
        freq = "T"
    elif request.param in "months":
        freq = "MS"
    elif request.param == "seasons":
        freq = "QS"
    elif request.param == "years":
        freq = "YS"
    else:
        freq = request.param[0].upper()
    # create initialized
    init = xr.cftime_range(START, freq=freq, periods=NINITS)
    lead = np.arange(NLEADS)
    member = np.arange(NMEMBERS)
    initialized = xr.DataArray(
        np.random.rand(len(init), len(lead), len(member)),
        dims=["init", "lead", "member"],
        coords=[init, lead, member],
    ).to_dataset(name="var")
    initialized.lead.attrs["units"] = request.param

    # create observations
    time = xr.cftime_range(START, freq=freq, periods=NINITS + NLEADS)
    obs = xr.DataArray(
        np.random.rand(len(time)), dims=["time"], coords=[time]
    ).to_dataset(name="var")
    # climpred.PredictionEnsemble
    hindcast = HindcastEnsemble(initialized).add_observations(obs)
    return hindcast


def test_HindcastEnsemble_time_resolution_verify(HindcastEnsemble_time_resolution):
    """Test that HindcastEnsemble.verify() in any lead time resolution works."""
    assert (
        HindcastEnsemble_time_resolution.verify(**HindcastEnsemble_verify_kw)
        .notnull()
        .any()
    )


def test_PerfectModelEnsemble_time_resolution_verify(HindcastEnsemble_time_resolution):
    """Test that PerfectModelEnsemble.verify() in any lead time resolution works."""
    pm = PerfectModelEnsemble(HindcastEnsemble_time_resolution.get_initialized())
    assert pm.verify(**PerfectModelEnsemble_verify_kw).notnull().any()


import pandas as pd


@pytest.mark.parametrize(
    "lead_res", ["seconds", "minutes", "hours", "days", "pentads", "weeks"]
)
def test_HindcastEnsemble_lead_pdTimedelta(hind_da_initialized_1d, lead_res):
    """Test to see HindcastEnsemble can be initialized with lead as pd.Timedelta."""
    if lead_res == "pentads":
        n, freq = 5, "d"
    else:
        n, freq = 1, lead_res[0].lower()
    initialized = hind_da_initialized_1d

    initialized["lead"] = [
        pd.Timedelta(f"{i*n} {freq}") for i in initialized.lead.values
    ]
    hindcast = HindcastEnsemble(initialized)

    assert hindcast.get_initialized().lead.attrs["units"] == lead_res


def test_monthly_leads_real_example(hindcast_NMME_Nino34):
    skill = (
        hindcast_NMME_Nino34.isel(lead=[0, 1, 2])
        .sel(init=slice("2005", "2006"))
        .verify(
            metric="crps",
            comparison="m2o",
            dim=["init", "member"],
            alignment="same_inits",
        )
    )
    assert skill.to_array().notnull().all()


def test_daily_leads_real_example(hindcast_S2S_Germany):
    skill = (
        hindcast_S2S_Germany.isel(lead=[0, 1])
        .sel(init=slice("2005", "2006"))
        .verify(
            metric="crps",
            comparison="m2o",
            dim=["init", "member"],
            alignment="same_inits",
        )
    )
    assert skill.to_array().notnull().all()



import pytest
import xarray as xr
from IPython.display import display

from climpred.classes import HindcastEnsemble, PerfectModelEnsemble


@pytest.mark.parametrize("display_style", ("html", "text"))
def test_repr_PM(PM_da_initialized_1d, PM_da_control_1d, display_style):
    """Test html and text repr."""
    with xr.set_options(display_style=display_style):
        pm = PerfectModelEnsemble(PM_da_initialized_1d)
        display(pm)
        pm = pm.add_control(PM_da_control_1d)
        display(pm)
        pm = pm.generate_uninitialized()
        display(pm)


@pytest.mark.parametrize("display_style", ("html", "text"))
def test_repr_HC(
    hind_ds_initialized_1d,
    hist_ds_uninitialized_1d,
    observations_ds_1d,
    display_style,
):
    """Test html repr."""
    with xr.set_options(display_style=display_style):
        he = HindcastEnsemble(hind_ds_initialized_1d)
        display(he)
        he = he.add_uninitialized(hist_ds_uninitialized_1d)
        display(he)
        he = he.add_observations(observations_ds_1d)
        display(he)
        # no uninit
        he = HindcastEnsemble(hind_ds_initialized_1d)
        he = he.add_observations(observations_ds_1d)
        display(he)



import numpy as np
import pytest

from climpred.checks import (
    has_dataset,
    has_dims,
    has_min_len,
    has_valid_lead_units,
    is_in_list,
    is_xarray,
    match_initialized_dims,
    match_initialized_vars,
)
from climpred.constants import VALID_LEAD_UNITS
from climpred.exceptions import DatasetError, DimensionError, VariableError


@is_xarray(0)
def _arbitrary_ds_da_func(ds_da, *args, **kwargs):
    """Function for testing if checking the first item in arg list is ds/da."""
    return ds_da, args, kwargs


@is_xarray([0, 1])
def _arbitrary_two_xr_func(ds_da1, ds_da2, *args, **kwargs):
    """Function for testing if checking the first two items in arg list is ds/da."""
    return ds_da1, ds_da2, args, kwargs


@is_xarray([0, 2])
def _arbitrary_two_xr_func_random_loc(ds_da1, some_arg, ds_da2, **kwargs):
    """Function for testing if checking the 1st and 3rd item in arg list is ds/da."""
    return ds_da1, some_arg, ds_da2, kwargs


@is_xarray([0, "da", "other_da"])
def _arbitrary_three_xr_func_args_keys(ds, da=None, other_da=None, **kwargs):
    """Function for testing if checking the first in arg list and the
    keywords da/other_da is ds/da."""
    return ds, da, other_da, kwargs


def test_is_xarray_ds(ds1):
    """Test if checking the first item in arg list is ds."""
    ds, args, kwargs = _arbitrary_ds_da_func(ds1, "arg1", "arg2", kwarg1="kwarg1")
    assert (ds1 == ds).all()
    assert args == ("arg1", "arg2")
    assert kwargs == {"kwarg1": "kwarg1"}


def test_is_xarray_not_ds():
    """Test if checking the first item in arg list is not a ds/da, raise an error."""
    not_a_ds = "not_a_ds"
    with pytest.raises(IOError) as e:
        _arbitrary_ds_da_func(not_a_ds, "arg1", "arg2", kwarg1="kwarg1")
    assert "The input data is not an xarray" in str(e.value)


def test_is_xarray_da(da1):
    """Test if checking the first item in arg list is da."""
    da, args, kwargs = _arbitrary_ds_da_func(da1, "arg1", "arg2", kwarg1="kwarg1")
    assert (da1 == da).all()
    assert args == ("arg1", "arg2")
    assert kwargs == {"kwarg1": "kwarg1"}


def test_is_xarray_ds_da(ds1, da1):
    """Test if checking the first two items in arg list is ds/da."""
    ds, da, args, kwargs = _arbitrary_two_xr_func(
        ds1, da1, "arg1", kwarg1="kwarg1", kwarg2="kwarg2"
    )
    assert (ds1 == ds).all()
    assert (da1 == da).all()
    assert args == ("arg1",)
    assert kwargs == {"kwarg1": "kwarg1", "kwarg2": "kwarg2"}


def test_is_xarray_ds_da_random_loc(ds1, da1):
    """Test if checking the first and third items in arg list is ds/da."""
    ds, arg, da, kwargs = _arbitrary_two_xr_func_random_loc(
        ds1, "arg1", da1, kwarg1="kwarg1", kwarg2="kwarg2"
    )
    assert (ds1 == ds).all()
    assert (da1 == da).all()
    assert arg == "arg1"
    assert kwargs == {"kwarg1": "kwarg1", "kwarg2": "kwarg2"}


def test_is_xarray_ds_da_args_keys(ds1, da1, da2):
    """Test if checking the args and kwargs are ds/da."""
    ds, da, other_da, kwargs = _arbitrary_three_xr_func_args_keys(
        ds1, da=da1, other_da=da2, kwarg1="kwarg1"
    )
    assert (ds1 == ds).all()
    assert (da1 == da).all()
    assert (da2 == other_da).all()
    assert kwargs == {"kwarg1": "kwarg1"}


def test_is_xarray_ds_da_args_keys_not(ds1, da2):
    """Test if checking the args and kwargs are not ds/da, it raises an error."""
    not_a_da = np.array([0, 1, 2])
    with pytest.raises(IOError) as e:
        _arbitrary_three_xr_func_args_keys(
            ds1, da=not_a_da, other_da=da2, kwarg1="kwarg1"
        )
    assert "The input data is not an xarray" in str(e.value)


class _ArbitraryClass:
    @is_xarray(1)
    def __init__(self, xobj):
        pass


def test_is_xarray_class_not():
    """Function for testing if checking class init is ds/da, it raises an error."""
    with pytest.raises(IOError) as e:
        _ArbitraryClass("totally not a ds")
    assert "The input data is not an xarray" in str(e.value)


def test_has_dims_str(da1):
    """Test if check works for a string."""
    assert has_dims(da1, "x", "arbitrary")


def test_has_dims_list(da1):
    """Test if check works for a list."""
    # returns None if no errors
    assert has_dims(da1, ["x", "y"], "arbitrary")


def test_has_dims_str_fail(da1):
    """Test if check fails properly for a string."""
    with pytest.raises(DimensionError) as e:
        has_dims(da1, "z", "arbitrary")
    assert "Your arbitrary object must contain" in str(e.value)


def test_has_dims_list_fail(da1):
    """Test if check fails properly for a list."""
    with pytest.raises(DimensionError) as e:
        has_dims(da1, ["z"], "arbitrary")
    assert "Your arbitrary object must contain" in str(e.value)


def test_has_min_len_arr(da1):
    """Test if check works for min len."""
    assert has_min_len(da1.values, 2, "arbitrary")


def test_has_min_len_fail(da1):
    """Test if check fails properly."""
    with pytest.raises(DimensionError) as e:
        has_min_len(da1.values, 5, "arbitrary")
    assert "Your arbitrary array must be at least" in str(e.value)


def test_has_dataset():
    """Test if check works for a non-empty list."""
    obj = [5]
    assert has_dataset(obj, "list", "something")


def test_has_dataset_fail():
    """Test if check works to fail for an empty list."""
    obj = []
    with pytest.raises(DatasetError) as e:
        has_dataset(obj, "test", "something")
    assert "You need to add at least one test dataset" in str(e.value)


def test_match_initialized_dims(da1, da2):
    """Test if check works if both da has the proper dims."""
    assert match_initialized_dims(da1.rename({"y": "init"}), da2.rename({"y": "time"}))


def test_match_initialized_dims_fail(da1, da2):
    """Test if check works if the da does not have the proper dims."""
    with pytest.raises(DimensionError) as e:
        match_initialized_dims(da1.rename({"y": "init"}), da2.rename({"y": "not_time"}))
    assert "Verification contains more dimensions than initialized" in str(e.value)


def test_match_initialized_vars(ds1, ds2):
    """Test if check works if both have the same variables."""
    assert match_initialized_vars(ds1, ds2)


def test_match_initialized_vars_fail(ds1, ds2):
    """Test if check works if both do not have the same variables."""
    with pytest.raises(VariableError) as e:
        match_initialized_vars(ds1, ds2.rename({"air": "tmp"}))
    assert "Please provide a Dataset/DataArray with at least" in str(e.value)


def test_is_in_list():
    """Test if check works if key is in dict."""
    some_list = ["key"]
    assert is_in_list("key", some_list, "metric")


def test_is_in_list_fail():
    """Test if check works if key is not in dict."""
    some_list = ["key"]
    with pytest.raises(KeyError) as e:
        is_in_list("not_key", some_list, "metric")
    assert "Specify metric from" in str(e.value)


@pytest.mark.parametrize("lead_units", VALID_LEAD_UNITS)
def test_valid_lead_units(da_lead, lead_units):
    """Test that lead units check passes with appropriate lead units."""
    da_lead["lead"].attrs["units"] = lead_units
    assert has_valid_lead_units(da_lead)


def test_valid_lead_units_no_units(da_lead):
    """Test that valid lead units check breaks if there are no units."""
    with pytest.raises(AttributeError):
        has_valid_lead_units(da_lead)


def test_valid_lead_units_invalid_units(da_lead):
    """Test that valid lead units check breaks if invalid units provided."""
    da_lead["lead"].attrs["units"] = "dummy"
    with pytest.raises(AttributeError):
        has_valid_lead_units(da_lead)


@pytest.mark.parametrize("lead_units", VALID_LEAD_UNITS)
def test_nonplural_lead_units_works(da_lead, lead_units):
    """Test that non-plural lead units work on lead units check."""
    da_lead["lead"].attrs["units"] = lead_units[:-1]
    with pytest.warns(UserWarning) as record:
        has_valid_lead_units(da_lead)
    expected = f'The letter "s" was appended to the lead units; now {lead_units}.'
    assert record[0].message.args[0] == expected



import numpy as np
import xarray as xr

from .checks import is_xarray

try:
    import xesmf as xe
except ImportError:
    xe = None


@is_xarray(0)
def spatial_smoothing_xesmf(
    ds,
    d_lon_lat_kws={"lon": 5, "lat": 5},
    method="bilinear",
    periodic=False,
    filename=None,
    reuse_weights=False,
    tsmooth_kws=None,
    how=None,
):
    """
    Quick regridding function. Adapted from
    https://github.com/JiaweiZhuang/xESMF/pull/27/files#diff-b537ef68c98c2ec11e64e4803fe4a113R105.

    Args:
        ds (xarray-object): Contain input and output grid coordinates.
            Look for coordinates ``lon``, ``lat``, and optionally ``lon_b``,
            ``lat_b`` for conservative method. Also any coordinate which is C/F
            compliant, .i.e. standard_name in ['longitude', 'latitude'] is allowed.
            Shape can be 1D (Nlon,) and (Nlat,) for rectilinear grids,
            or 2D (Ny, Nx) for general curvilinear grids.
            Shape of bounds should be (N+1,) or (Ny+1, Nx+1).
        d_lon_lat_kws (dict): optional
            Longitude/Latitude step size (grid resolution); if not provided,
            lon will equal 5 and lat will equal lon
            (optional)
        method (str): Regridding method. Options are:
            - 'bilinear'
            - 'conservative', **need grid corner information**
            - 'patch'
            - 'nearest_s2d'
            - 'nearest_d2s'
        periodic (bool): Periodic in longitude? Default to False. optional
            Only useful for global grids with non-conservative regridding.
            Will be forced to False for conservative regridding.
        filename (str): Name for the weight file. (optional)
            The default naming scheme is:
                {method}_{Ny_in}x{Nx_in}_{Ny_out}x{Nx_out}.nc
                e.g. bilinear_400x600_300x400.nc
        reuse_weights (bool) Whether to read existing weight file to save
            computing time. False by default. (optional)
        tsmooth_kws (None): leads nowhere but consistent with `temporal_smoothing`.
        how (None): leads nowhere but consistent with `temporal_smoothing`.

        Returns:
            ds (xarray.object) regridded
    """

    if xe is None:
        raise ImportError(
            "xesmf is not installed; see"
            "https://pangeo-xesmf.readthedocs.io/en/latest/installation.html"
        )

    def _regrid_it(da, d_lon, d_lat, **kwargs):
        """
        Global 2D rectilinear grid centers and bounds

        Args:
            da (xarray.DataArray): Contain input and output grid coords.
                Look for variables ``lon``, ``lat``, ``lon_b``, ``lat_b`` for
                conservative method, and ``TLAT``, ``TLON`` for CESM POP grid
                Shape can be 1D (Nlon,) and (Nlat,) for rectilinear grids,
                or 2D (Ny, Nx) for general curvilinear grids.
                Shape of bounds should be (N+1,) or (Ny+1, Nx+1).
            d_lon (float): Longitude step size, i.e. grid resolution
            d_lat (float): Latitude step size, i.e. grid resolution
        Returns:
            da : xarray DataArray with coordinate values
        """

        if "lon" in da.coords:
            lon = da.lon
        else:
            try:
                lon = da.cf["longitude"]
            except KeyError:
                raise KeyError(
                    "Could not find `lon` as coordinate or any C/F compliant `latitude` coordinate, see https://pangeo-xesmf.readthedocs.io and https://cf-xarray.readthedocs.io"
                )

        if "lat" in da.coords:
            lat = da.lat
        else:
            try:
                lat = da.cf["latitude"]
            except KeyError:
                raise KeyError(
                    "C/F compliant or `lat` as coordinate, see https://pangeo-xesmf.readthedocs.io"
                )

        grid_out = xr.Dataset(
            {
                "lat": (["lat"], np.arange(lat.min(), lat.max() + d_lat, d_lat)),
                "lon": (["lon"], np.arange(lon.min(), lon.max() + d_lon, d_lon)),
            }
        )
        regridder = xe.Regridder(da, grid_out, **kwargs)
        return regridder(da)

    # check if lon or/and lat missing
    if ("lon" in d_lon_lat_kws) and ("lat" in d_lon_lat_kws):
        pass
    elif ("lon" not in d_lon_lat_kws) and ("lat" in d_lon_lat_kws):
        d_lon_lat_kws["lon"] = d_lon_lat_kws["lat"]
    elif ("lat" not in d_lon_lat_kws) and ("lon" in d_lon_lat_kws):
        d_lon_lat_kws["lat"] = d_lon_lat_kws["lon"]
    else:
        raise ValueError("please provide either `lon` or/and `lat` in d_lon_lat_kws.")

    kwargs = {
        "d_lon": d_lon_lat_kws["lon"],
        "d_lat": d_lon_lat_kws["lat"],
        "method": method,
        "periodic": periodic,
        "filename": filename,
        "reuse_weights": reuse_weights,
    }

    ds = _regrid_it(ds, **kwargs)

    return ds


@is_xarray(0)
def temporal_smoothing(ds, tsmooth_kws=None, how="mean", d_lon_lat_kws=None):
    """Apply temporal smoothing by creating rolling smooth-timestep means.

    Reference:
    * Goddard, L., A. Kumar, A. Solomon, D. Smith, G. Boer, P.
     Gonzalez, V. Kharin, et al. “A Verification Framework for
     Interannual - to - Decadal Predictions Experiments.” Climate
     Dynamics 40, no. 1–2 (January 1, 2013): 245–72.
     https://doi.org/10/f4jjvf.

    Args:
        ds(xr.object): input.
        tsmooth_kws(dict): length of smoothing of timesteps.
            Defaults to {'time': 4} (see Goddard et al. 2013).
        how(str): aggregation type for smoothing. default: 'mean'
        d_lon_lat_kws (None): leads nowhere but consistent with
            `spatial_smoothing_xesmf`.

    Returns:
        ds_smoothed(xr.object): input with `smooth` timesteps less
            and labeling '1-(smooth-1)', '...', ... .

    """
    # unpack dict
    if not isinstance(tsmooth_kws, dict):
        raise ValueError(
            "Please provide tsmooth_kws as dict, found ", type(tsmooth_kws)
        )
    if not ("time" in tsmooth_kws or "lead" in tsmooth_kws):
        raise ValueError(
            'tsmooth_kws doesnt contain a time dimension \
            (either "lead" or "time").',
            tsmooth_kws,
        )
    smooth = list(tsmooth_kws.values())[0]
    if smooth == 1:
        return ds
    dim = list(tsmooth_kws.keys())[0]
    # fix to smooth either lead or time depending
    time_dims = ["time", "lead"]
    if dim not in ds.dims:
        time_dims.remove(dim)
        dim = time_dims[0]
        tsmooth_kws = {dim: smooth}
    # aggreate based on how
    ds_smoothed = getattr(ds.rolling(tsmooth_kws, center=False), how)()
    # remove first all-nans
    ds_smoothed = ds_smoothed.isel({dim: slice(smooth - 1, None)})
    ds_smoothed[dim] = ds.isel({dim: slice(None, -smooth + 1)})[dim]
    return ds_smoothed


def _reset_temporal_axis(ds_smoothed, tsmooth_kws, dim="lead", set_lead_center=True):
    """Reduce and reset temporal axis. See temporal_smoothing(). Should be
    used after calculation of skill to maintain readable labels for skill
    computation.

    Args:
        ds_smoothed (xarray object): Smoothed dataset.
        tsmooth_kws (dict): Keywords smoothing is performed over.
        dim (str): Dimension smoothing is performed over. Defaults to 'lead'.
        set_center (bool): Whether to set new coord `{dim}_center`.
            Defaults to True.

    Returns:
        Smoothed Dataset with updated labels for smoothed temporal dimension.
    """
    # bugfix: actually tsmooth_kws should only dict
    if tsmooth_kws is None or callable(tsmooth_kws):
        return ds_smoothed
    if not ("time" in tsmooth_kws.keys() or "lead" in tsmooth_kws.keys()):
        raise ValueError("tsmooth_kws does not contain a time dimension.", tsmooth_kws)
    for c in ["time", "lead"]:
        if c in tsmooth_kws.keys():
            smooth = tsmooth_kws[c]
    ds_smoothed[dim] = [f"{t}-{t + smooth - 1}" for t in ds_smoothed[dim].values]
    if set_lead_center:
        _set_center_coord(ds_smoothed, dim)
    return ds_smoothed


def _set_center_coord(ds, dim="lead"):
    """Set lead_center as a new coordinate."""
    new_dim = []
    old_dim = ds[dim].values
    for i in old_dim:
        new_dim.append(eval(i.replace("-", "+")) / 2)
    new_dim = np.array(new_dim)
    ds.coords[f"{dim}_center"] = (dim, new_dim)
    return ds


@is_xarray(0)
def smooth_goddard_2013(
    ds,
    tsmooth_kws={"lead": 4},
    d_lon_lat_kws={"lon": 5, "lat": 5},
    how="mean",
    **xesmf_kwargs,
):
    """Wrapper to smooth as suggested by Goddard et al. 2013:
        - 4-year composites
        - 5x5 degree regridding

    Reference:
    * Goddard, L., A. Kumar, A. Solomon, D. Smith, G. Boer, P.
        Gonzalez, V. Kharin, et al. “A Verification Framework for
        Interannual - to - Decadal Predictions Experiments.” Climate
        Dynamics 40, no. 1–2 (January 1, 2013): 245–72.
        https: // doi.org / 10 / f4jjvf.

    Args:
        ds(xr.object): input.
        tsmooth_kws(dict): length of smoothing of timesteps (applies to ``lead``
                          in forecast and ``time`` in verification data).
                          Default: {'time': 4} (see Goddard et al. 2013).
        d_lon_lat_kws (dict): target grid for regridding.
                              Default: {'lon':5 , 'lat': 5}
        how(str): aggregation type for smoothing. default: 'mean'
        **xesmf_kwargs (kwargs): kwargs passed to `spatial_smoothing_xesmf`.

    Returns:
        ds_smoothed_regridded (xr.object): input with `smooth` timesteps less
                                           and labeling '1-(smooth-1)', '...' .

    """
    # first temporal smoothing
    ds_smoothed = temporal_smoothing(ds, tsmooth_kws=tsmooth_kws)
    ds_smoothed_regridded = spatial_smoothing_xesmf(
        ds_smoothed, d_lon_lat_kws=d_lon_lat_kws, **xesmf_kwargs
    )
    return ds_smoothed_regridded



import warnings
from collections import OrderedDict

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from xarray.coding.times import infer_calendar_name

from .checks import DimensionError
from .constants import CLIMPRED_DIMS
from .metrics import ALL_METRICS, PROBABILISTIC_METRICS
from .utils import get_lead_cftime_shift_args, get_metric_class, shift_cftime_index


def plot_relative_entropy(rel_ent, rel_ent_threshold=None, **kwargs):
    """
    Plot relative entropy results.

    Args:
        rel_ent (xr.Dataset): relative entropy from compute_relative_entropy
        rel_ent_threshold (xr.Dataset): threshold from
                                          bootstrap_relative_entropy
        **kwargs: for plt.subplots( **kwargs)

    """
    colors = ["royalblue", "indianred", "goldenrod"]
    _, ax = plt.subplots(ncols=3, **kwargs)

    for i, dim in enumerate(["R", "S", "D"]):
        m = rel_ent[dim].median("init")
        std = rel_ent[dim].std("init")
        ax[i].plot(
            rel_ent.lead,
            rel_ent[dim].to_dataframe().unstack(0),
            c="gray",
            label="individual initializations",
            linewidth=0.5,
            alpha=0.5,
        )
        ax[i].plot(rel_ent.lead, m, c=colors[i], label=dim, linewidth=2.5)
        ax[i].plot(
            rel_ent.lead,
            (m - std),
            c=colors[i],
            label=dim + " median +/- std",
            linewidth=2.5,
            ls="--",
        )
        ax[i].plot(
            rel_ent.lead,
            (m + std),
            c=colors[i],
            label="",
            linewidth=2.5,
            ls="--",
        )
        if rel_ent_threshold is not None:
            ax[i].axhline(
                y=rel_ent_threshold[dim].values,
                label="bootstrapped threshold",
                c="gray",
                ls="--",
            )
        handles, labels = ax[i].get_legend_handles_labels()
        by_label = OrderedDict(zip(labels, handles))
        ax[i].legend(by_label.values(), by_label.keys(), frameon=False)
        ax[i].set_xlabel("Lead")
    ax[0].set_title("Relative Entropy")
    ax[1].set_title("Signal")
    ax[2].set_title("Dispersion")
    ax[0].set_ylabel("Relative Entropy [ ]")
    ax[0].set_ylim(bottom=0)
    return ax


def plot_bootstrapped_skill_over_leadyear(
    bootstrapped,
    ax=None,
    color_initialized="indianred",
    color_uninitialized="steelblue",
    color_persistence="gray",
    color_climatology="tan",
    capsize=4,
    fontsize=8,
    figsize=(10, 4),
    fmt="--o",
):
    """
    Plot Ensemble Prediction skill as in Li et al. 2016 Fig.3a-c.

    Args:
        bootstrapped (xr.DataArray or xr.Dataset with one variable):
            from PredictionEnsembleEnsemble.bootstrap() or HindcastEnsemble.bootstrap()

        ax (plt.axes): plot on ax. Defaults to None.

    Returns:
        ax

    Reference:
      * Li, Hongmei, Tatiana Ilyina, Wolfgang A. Müller, and Frank
            Sienz. “Decadal Predictions of the North Atlantic CO2
            Uptake.” Nature Communications 7 (March 30, 2016): 11076.
            https://doi.org/10/f8wkrs.

    """
    if isinstance(bootstrapped, xr.Dataset):
        var = list(bootstrapped.data_vars)
        if len(var) > 1:
            raise ValueError(
                "Please provide only xr.Dataset with one variable or xr.DataArray."
            )
        # copy attributes to xr.DataArray
        elif len(var) == 1:
            var = var[0]
            attrs = bootstrapped.attrs
            bootstrapped = bootstrapped[var]
            bootstrapped.attrs = attrs

    assert isinstance(bootstrapped, xr.DataArray)
    reference = list(bootstrapped.drop_sel(skill="initialized").coords["skill"].values)

    sig = bootstrapped.attrs["confidence_interval_levels"].split("-")
    sig = int(100 * (float(sig[0]) - float(sig[1])))
    pers_sig = sig

    init_skill = bootstrapped.sel(skill="initialized", results="verify skill")
    init_ci = bootstrapped.sel(
        skill="initialized", results=["low_ci", "high_ci"]
    ).rename({"results": "quantile"})

    if pers_sig != sig:
        raise NotImplementedError("pers_sig != sig not implemented yet.")

    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
    # plot init
    ax.errorbar(
        init_skill.lead,
        init_skill,
        yerr=[
            init_skill - init_ci.isel(quantile=0),
            init_ci.isel(quantile=1) - init_skill,
        ],
        fmt=fmt,
        capsize=capsize,
        c=color_initialized,
        label="initialized",
    )
    # plot references
    for r in reference:
        r_skill = bootstrapped.sel(skill=r, results="verify skill")
        if (r_skill == np.nan).all():
            warnings.warn(f"Found only NaNs in {r} verify skill and skipped.")
            continue
        p_r_over_init = bootstrapped.sel(skill=r, results="p")
        r_ci = bootstrapped.sel(skill=r, results=["low_ci", "high_ci"]).rename(
            {"results": "quantile"}
        )
        c = eval(f"color_{r}")
        # add p values over all reference skills
        for t in init_skill.lead.values:
            ax.text(
                r_skill.lead.sel(lead=t),
                r_ci.isel(quantile=0).sel(lead=t).values,
                "%.2f" % float(p_r_over_init.sel(lead=t).values),
                horizontalalignment="center",
                verticalalignment="bottom",
                fontsize=fontsize,
                color=c,
            )
        yerr = [
            r_skill - r_ci.isel(quantile=0),
            r_ci.isel(quantile=1) - r_skill,
        ]
        x = r_skill.lead
        ax.errorbar(
            x,
            r_skill,
            yerr=yerr,
            fmt=fmt,
            capsize=capsize,
            c=c,
            label=r,
        )

    ax.xaxis.set_ticks(bootstrapped.lead.values)
    ax.legend(frameon=False, title=f"skill with {sig}% confidence interval:")
    ax.set_xlabel(f"Lead time [{bootstrapped.lead.attrs['units']}]")
    ax.set_ylabel(get_metric_class(bootstrapped.attrs["metric"], ALL_METRICS).long_name)
    return ax


def _check_only_climpred_dims(pe):
    """Warns if dimensions other than `CLIMPRED_DIMS` are in `PredictionEnsemble`."""
    additional_dims = set(pe.get_initialized().dims) - set(CLIMPRED_DIMS)
    if len(additional_dims) != 0:
        raise DimensionError(
            f"{type(pe.__name__)}.plot() does not allow dimensions other "
            f"than {CLIMPRED_DIMS}, found {additional_dims}. "
            f"Please use .mean({additional_dims}) "
            f"or .isel() before plot."
        )


def plot_lead_timeseries_hindcast(
    he, variable=None, ax=None, show_members=False, cmap="viridis"
):
    """Plot datasets from HindcastEnsemble.

    Args:
        he (HindcastEnsemble): HindcastEnsemble.
        variable (str or None): `variable` to plot. Defaults to the first in data_vars.
        ax (plt.axes): Axis to use in plotting. By default, creates a new axis.
        show_members (bool): whether to display all members individually.
            Defaults to False.
        cmap (str): Name of matplotlib-recognized colorbar. Defaults to 'viridis'.

    Returns:
        ax: plt.axes

    """
    _check_only_climpred_dims(he)
    if variable is None:
        variable = list(he.get_initialized().data_vars)[0]
    hind = he.get_initialized()[variable]
    lead_freq = get_lead_cftime_shift_args(hind.lead.attrs["units"], 1)
    lead_freq = str(lead_freq[0]) + lead_freq[1]
    hist = he.get_uninitialized()
    if isinstance(hist, xr.Dataset):
        hist = hist[variable]
    obs = he._datasets["observations"]

    cmap = mpl.cm.get_cmap(cmap, hind.lead.size)
    if ax is None:
        _, ax = plt.subplots(figsize=(10, 4))
    if isinstance(hist, xr.DataArray):
        if "member" in hist.dims and not show_members:
            hist = hist.mean("member")
            member_alpha = 1
            lw = 2
        else:
            member_alpha = 0.4
            lw = 1
        hist.plot(
            ax=ax,
            lw=lw,
            hue="member",
            color="gray",
            alpha=member_alpha,
            label="uninitialized",
            zorder=hind.lead.size + 1,
        )

    for i, lead in enumerate(hind.lead.values):
        h = hind.sel(lead=lead).rename({"init": "time"})
        if not show_members and "member" in h.dims:
            h = h.mean("member")
            lead_alpha = 1
        else:
            lead_alpha = 0.5
        h["time"] = shift_cftime_index(h.time, "time", int(lead), lead_freq)
        h.plot(
            ax=ax,
            hue="member",
            color=cmap(i),
            label=f"initialized: lead={lead} {hind.lead.attrs['units'][:-1]}",
            alpha=lead_alpha,
            zorder=hind.lead.size - i,
        )

    if len(obs) > 0:
        if isinstance(obs, xr.Dataset):
            obs = obs[variable]
        obs.plot(
            ax=ax,
            color="k",
            lw=3,
            ls="-",
            label="observations",
            zorder=hind.lead.size + 2,
        )

    # show only one item per label in legend
    handles, labels = ax.get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    ax.legend(
        by_label.values(), by_label.keys(), loc="center left", bbox_to_anchor=(1, 0.5)
    )
    ax.set_title("")
    return ax


def plot_ensemble_perfect_model(
    pm, variable=None, ax=None, show_members=False, cmap="tab10"
):
    """Plot datasets from PerfectModelEnsemble.

    Args:
        pm (PerfectModelEnsemble): PerfectModelEnsemble.
        variable (str or None): `variable` to plot. Defaults to the first in data_vars.
        ax (plt.axes): Axis to use in plotting. By default, creates a new axis.
        show_members (bool): whether to display all members individually.
            Defaults to False.
        cmap (str): Name of matplotlib-recognized colorbar. Defaults to 'tab10'.

    Returns:
        ax: plt.axes

    """

    _check_only_climpred_dims(pm)
    if variable is None:
        variable = list(pm.get_initialized().data_vars)[0]
    initialized = pm.get_initialized()[variable]
    uninitialized = pm.get_uninitialized()
    if isinstance(uninitialized, xr.Dataset):
        uninitialized = uninitialized[variable]
        uninitialized_present = True
    else:
        uninitialized_present = False
    control = pm.get_control()
    if isinstance(control, xr.Dataset):
        control = control[variable]
    calendar = infer_calendar_name(initialized.init)
    lead_freq = get_lead_cftime_shift_args(initialized.lead.attrs["units"], 1)[1]

    control_color = "gray"

    if ax is None:
        _, ax = plt.subplots(figsize=(10, 4))

    cmap = mpl.cm.get_cmap(cmap, initialized.init.size)

    for ii, i in enumerate(initialized.init.values):
        dsi = initialized.sel(init=i).rename({"lead": "time"})
        if uninitialized_present:
            dsu = uninitialized.sel(init=i).rename({"lead": "time"})
        # convert lead time into cftime
        start_str = i.strftime()[:10]
        if initialized.lead.min() == 0:
            dsi["time"] = xr.cftime_range(
                start=start_str,
                freq=lead_freq,
                periods=dsi.time.size,
                calendar=calendar,
            )
        elif initialized.lead.min() == 1:
            dsi["time"] = xr.cftime_range(
                start=start_str,
                freq=lead_freq,
                periods=dsi.time.size,
                calendar=calendar,
            )
            dsi["time"] = shift_cftime_index(dsi.time, "time", 1, lead_freq)
        if uninitialized_present:
            dsu["time"] = dsi["time"]
        if not show_members:
            dsi = dsi.mean("member")
            if uninitialized_present:
                dsu = dsu.mean("member")
            member_alpha = 1
            lw = 2
            labelstr = "ensemble mean"
        else:
            member_alpha = 0.5
            lw = 1
            labelstr = "members"
            # plot ensemble mean, first white then color to highlight ensemble mean
            if uninitialized_present:
                dsu.mean("member").plot(ax=ax, color="white", lw=3, zorder=8, alpha=0.6)
                dsu.mean("member").plot(
                    ax=ax, color=control_color, lw=2, zorder=9, alpha=0.6
                )
            # plot ensemble mean, first white then color to highlight ensemble mean
            dsi.mean("member").plot(ax=ax, color="white", lw=3, zorder=10)
            dsi.mean("member").plot(ax=ax, color=cmap(ii), lw=2, zorder=11)
        dsi.plot(
            ax=ax,
            hue="member",
            color=cmap(ii),
            alpha=member_alpha,
            lw=lw,
            label=labelstr,
        )
        if uninitialized_present:
            dsu.plot(
                ax=ax,
                hue="member",
                color=control_color,
                alpha=member_alpha / 2,
                lw=lw,
                label="uninitialized " + labelstr,
            )

    if isinstance(control, xr.DataArray):
        control.plot(ax=ax, color=control_color, label="control")

    # show only one item per label in legend
    handles, labels = ax.get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    ax.legend(by_label.values(), by_label.keys())
    ax.set_title(" ")
    return ax



import numpy as np
import pytest
import xarray as xr
from xarray.testing import assert_equal

from climpred import HindcastEnsemble, PerfectModelEnsemble
from climpred.classes import PredictionEnsemble
from climpred.constants import CF_LONG_NAMES, CF_STANDARD_NAMES, CLIMPRED_DIMS
from climpred.exceptions import VariableError
from climpred.testing import assert_PredictionEnsemble, check_dataset_dims_and_data_vars

xr.set_options(display_style="text")

ALLOWED_TYPES_FOR_MATH_OPERATORS = [
    int,
    float,
    np.ndarray,
    xr.DataArray,
    xr.Dataset,
    PredictionEnsemble,
]

OTHER_ALLOWED_HINDCAST = [
    2,
    2.0,
    np.array(2),
    xr.DataArray(2),
    xr.DataArray(2).chunk(),
    xr.Dataset({"SST": 2}),
    xr.Dataset({"SST": 2}).chunk(),
]
OTHER_ALLOWED_PM = [
    2,
    2.0,
    np.array(2),
    xr.DataArray(2),
    xr.DataArray(2).chunk(),
    xr.Dataset({"tos": 2}),
    xr.Dataset({"tos": 2}).chunk(),
]
OTHER_ALLOWED_IDS = [
    "int",
    "float",
    "np.array",
    "xr.DataArray",
    "chunked xr.DataArray",
    "xr.Dataset",
    "chunked xr.Dataset",
]

OTHER_NOT_ALLOWED = ["2", list(), dict(), set()]
OTHER_NOT_ALLOWED_IDS = ["str", "list", "dict", "set"]

MATH_OPERATORS = ["add", "sub", "mul", "div"]


def add(a, b):
    return a + b


def sub(a, b):
    return a - b


def mul(a, b):
    return a * b


def div(a, b):
    return a / b


def gen_error_str(he, operator, other):
    print(type(operator))
    OPERATOR_STR = {
        "add": "+",
        "sub": "-",
        "mul": "*",
        "div": "/",
    }
    return f"Cannot use {type(he)} {OPERATOR_STR[operator]} {type(other)}"


# HindcastEnsemble
@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
@pytest.mark.parametrize("other", OTHER_ALLOWED_HINDCAST, ids=OTHER_ALLOWED_IDS)
def test_hindcastEnsemble_plus_defined(
    hind_ds_initialized_1d,
    hist_ds_uninitialized_1d,
    observations_ds_1d,
    other,
    operator,
):
    """Test that HindcastEnsemble math operator (+-*/) other works correctly for
    allowed other types."""
    he = HindcastEnsemble(hind_ds_initialized_1d)
    he = he.add_uninitialized(hist_ds_uninitialized_1d)
    he = he.add_observations(observations_ds_1d)
    operator = eval(operator)
    he2 = operator(he, other)
    for dataset in he._datasets:
        if he._datasets[dataset]:
            print("check", dataset)
            assert_equal(he2._datasets[dataset], operator(he._datasets[dataset], other))
            # check same dims and data_vars as before
            check_dataset_dims_and_data_vars(he, he2, dataset)


@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
def test_hindcastEnsemble_plus_hindcastEnsemble(
    hind_ds_initialized_1d, hist_ds_uninitialized_1d, observations_ds_1d, operator
):
    """Test that HindcastEnsemble math operator (+-*/) HindcastEnsemble works
    correctly."""
    he = HindcastEnsemble(hind_ds_initialized_1d)
    he = he.add_uninitialized(hist_ds_uninitialized_1d)
    he = he.add_observations(observations_ds_1d)
    other = he.mean("init")
    operator = eval(operator)
    he2 = operator(he, other)
    for dataset in he._datasets:
        if he._datasets[dataset]:
            print("check", dataset)
            assert_equal(
                he2._datasets[dataset],
                operator(he._datasets[dataset], other._datasets[dataset]),
            )
            # check same dims and data_vars as before
            check_dataset_dims_and_data_vars(he, he2, dataset)


@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
@pytest.mark.parametrize("other", OTHER_NOT_ALLOWED, ids=OTHER_NOT_ALLOWED_IDS)
def test_hindcastEnsemble_plus_not_defined(hind_ds_initialized_1d, other, operator):
    """Test that HindcastEnsemble math operator (+-*/) other raises error for
    non-defined others."""
    he = HindcastEnsemble(hind_ds_initialized_1d)
    error_str = gen_error_str(he, operator, other)
    operator = eval(operator)
    with pytest.raises(TypeError) as excinfo:
        operator(he, other)
    assert f"{error_str} because type {type(other)} not supported" in str(excinfo.value)


@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
@pytest.mark.parametrize("other", OTHER_ALLOWED_PM[-2:], ids=OTHER_ALLOWED_IDS[-2:])
def test_hindcastEnsemble_plus_Dataset_different_name(
    hind_ds_initialized_1d, other, operator
):
    """Test that HindcastEnsemble math operator (+-*/) other raises error for
    non-defined others."""

    he = HindcastEnsemble(hind_ds_initialized_1d)
    error_str = gen_error_str(he, operator, other)
    operator = eval(operator)
    with pytest.raises(VariableError) as excinfo:
        operator(he, other)
    assert f"{error_str} with new `data_vars`" in str(excinfo.value)


@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
def test_hindcastEnsemble_plus_broadcast(hind_ds_initialized_3d, operator):
    """Test that HindcastEnsemble math operator (+-*/) other also broadcasts
    correctly."""
    he = HindcastEnsemble(hind_ds_initialized_3d)
    operator = eval(operator)
    # minimal adding an offset or like multiplying area
    he2 = operator(
        he, xr.ones_like(hind_ds_initialized_3d.isel(init=1, lead=1, drop=True))
    )
    he3 = operator(he, 1)
    assert_PredictionEnsemble(he2, he3)


def test_hindcastEnsemble_operator_different_datasets(
    hind_ds_initialized_1d, observations_ds_1d
):
    """Test that HindcastEnsemble math operator (+-*/) on different datasets attached to HindcastEnsemble."""
    he = HindcastEnsemble(hind_ds_initialized_1d)
    he = he.add_observations(observations_ds_1d)
    he2 = HindcastEnsemble(hind_ds_initialized_1d)
    assert not (he2 - he).equals(he2)
    assert not (he - he2).equals(he)


def test_HindcastEnsemble_area_weighted_mean(hind_ds_initialized_3d):
    """Test area weighted mean HindcastEnsemble."""
    he = HindcastEnsemble(hind_ds_initialized_3d)
    # fake area
    area = hind_ds_initialized_3d["TAREA"]
    spatial_dims = [d for d in hind_ds_initialized_3d.dims if d not in CLIMPRED_DIMS]
    # PredictionEnsemble doesnt like other data_vars
    he_self_spatial_mean = (he * area).sum(spatial_dims) / area.sum()
    # weighted requires Dataset
    area = area.to_dataset(name="area")
    he_xr_spatial_mean = he.weighted(area).mean(spatial_dims)
    assert_PredictionEnsemble(
        he_self_spatial_mean, he_xr_spatial_mean, how="allclose", rtol=0.03, atol=0.05
    )


# basically all copied
# PerfectModelEnsemble
@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
@pytest.mark.parametrize("other", OTHER_ALLOWED_PM, ids=OTHER_ALLOWED_IDS)
def test_PerfectModelEnsemble_plus_defined(
    PM_ds_initialized_1d, PM_ds_control_1d, other, operator
):
    """Test that PerfectModelEnsemble math operator (+-*/) other works correctly for
    allowed other types."""
    he = PerfectModelEnsemble(PM_ds_initialized_1d)
    he = he.add_control(PM_ds_control_1d)
    operator = eval(operator)
    he2 = operator(he, other)
    for dataset in he._datasets:
        if he._datasets[dataset]:
            print("check", dataset)
            assert_equal(he2._datasets[dataset], operator(he._datasets[dataset], other))
            # check same dims and data_vars as before
            check_dataset_dims_and_data_vars(he, he2, dataset)


@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
def test_PerfectModelEnsemble_plus_PerfectModelEnsemble(
    PM_ds_initialized_1d, PM_ds_control_1d, operator
):
    """Test that PerfectModelEnsemble math operator (+-*/) PerfectModelEnsemble works
    correctly."""
    he = PerfectModelEnsemble(PM_ds_initialized_1d)
    he = he.add_control(PM_ds_control_1d)
    other = he.mean("init")
    operator = eval(operator)
    he2 = operator(he, other)
    for dataset in he._datasets:
        if he._datasets[dataset]:
            print("check", dataset)
            assert_equal(
                he2._datasets[dataset],
                operator(he._datasets[dataset], other._datasets[dataset]),
            )
            # check same dims and data_vars as before
            check_dataset_dims_and_data_vars(he, he2, dataset)


@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
@pytest.mark.parametrize("other", OTHER_NOT_ALLOWED, ids=OTHER_NOT_ALLOWED_IDS)
def test_PerfectModelEnsemble_plus_not_defined(PM_ds_initialized_1d, other, operator):
    """Test that PerfectModelEnsemble math operator (+-*/) other raises error for
    non-defined others."""
    he = PerfectModelEnsemble(PM_ds_initialized_1d)
    error_str = gen_error_str(he, operator, other)
    operator = eval(operator)
    with pytest.raises(TypeError) as excinfo:
        operator(he, other)
    assert f"{error_str} because type {type(other)} not supported" in str(excinfo.value)


@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
@pytest.mark.parametrize(
    "other", OTHER_ALLOWED_HINDCAST[-2:], ids=OTHER_ALLOWED_IDS[-2:]
)
def test_PerfectModelEnsemble_plus_Dataset_different_name(
    PM_ds_initialized_1d, other, operator
):
    """Test that PerfectModelEnsemble math operator (+-*/) other raises error for
    Dataset with other dims and/or variables."""
    he = PerfectModelEnsemble(PM_ds_initialized_1d)
    error_str = gen_error_str(he, operator, other)
    operator = eval(operator)
    with pytest.raises(VariableError) as excinfo:
        operator(he, other)
    assert f"{error_str} with new `data_vars`" in str(excinfo.value)


@pytest.mark.parametrize("operator", MATH_OPERATORS, ids=MATH_OPERATORS)
def test_PerfectModelEnsemble_plus_broadcast(PM_ds_initialized_3d, operator):
    """Test that PerfectModelEnsemble math operator (+-*/) other also broadcasts
    correctly."""
    he = PerfectModelEnsemble(PM_ds_initialized_3d)
    operator = eval(operator)
    # minimal adding an offset or like multiplying area
    he2 = operator(
        he, xr.ones_like(PM_ds_initialized_3d.isel(init=1, lead=1, drop=True))
    )
    he3 = operator(he, 1)
    assert_PredictionEnsemble(he2, he3)


def test_PerfectModelEnsemble_area_weighted_mean(PM_ds_initialized_3d):
    """Test area weighted mean PerfectModelEnsemble."""
    he = PerfectModelEnsemble(PM_ds_initialized_3d)
    # fake area
    area = np.cos(PM_ds_initialized_3d.lat) + 1
    spatial_dims = [d for d in PM_ds_initialized_3d.dims if d not in CLIMPRED_DIMS]
    # PredictionEnsemble doesnt like other data_vars
    he_self_spatial_mean = (he * area).sum(spatial_dims) / area.sum()
    # weighted requires Dataset
    area = area.to_dataset(name="area")
    he_xr_spatial_mean = he.weighted(area).mean(spatial_dims)
    assert_PredictionEnsemble(
        he_self_spatial_mean, he_xr_spatial_mean, how="allclose", rtol=0.03, atol=0.05
    )


@pytest.mark.parametrize("varlist", [["tos", "sos"], ["AMO"], "AMO"])
def test_subset_getitem_datavariables(
    perfectModelEnsemble_3v_initialized_control_1d, varlist
):
    """Test variable subselection from __getitem__."""
    pm = perfectModelEnsemble_3v_initialized_control_1d
    all_datavars = list(pm.get_initialized().data_vars)
    pm_subset = pm[varlist]
    if isinstance(varlist, str):
        varlist = [varlist]
    # test that varlist is present
    for var in varlist:
        assert var in pm_subset.get_initialized().data_vars
    # test that others are not present anymore
    for var in all_datavars:
        if var not in varlist:
            assert var not in pm_subset.get_initialized().data_vars


@pytest.mark.parametrize("equal", [True, False])
def test_eq_ne(perfectModelEnsemble_3v_initialized_control_1d, equal):
    xr.set_options(display_style="text")
    pm = perfectModelEnsemble_3v_initialized_control_1d
    if equal:
        pm2 = pm
        assert isinstance(pm2, PredictionEnsemble)
        print(pm, pm2)
        print("expect: True, False")
        print(pm == pm2)
        print(pm != pm2)
        assert pm == pm2
        assert isinstance(pm == pm2, bool)
        assert not (pm != pm2)
        assert isinstance(pm != pm2, bool)
        # assert False
    else:
        pm2 = pm * 1.6
        assert isinstance(pm2, PredictionEnsemble)
        print(pm, pm2)
        print("expect: True, False")
        print(pm != pm2)
        print(pm == pm2)
        assert not (pm == pm2)
        assert isinstance(pm == pm2, bool)
        assert pm != pm2
    assert isinstance(pm != pm2, bool)


pe = [
    pytest.lazy_fixture("hindcast_hist_obs_1d"),
    pytest.lazy_fixture("perfectModelEnsemble_initialized_control"),
]
pe_ids = ["HindcastEnsemble", "PerfectModelEnsemble"]


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_data_vars(pe):
    assert isinstance(pe.data_vars, xr.core.dataset.DataVariables)
    assert list(pe.data_vars) == list(pe.get_initialized().data_vars)
    assert len(pe) == len(pe.data_vars)


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_delitem(pe):
    v = list(pe.data_vars)[0]
    del pe[v]
    assert list(pe.data_vars) == []


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_nbytes(pe):
    assert pe.nbytes > pe.get_initialized().nbytes


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_coords(pe):
    assert isinstance(pe.coords, xr.core.coordinates.DatasetCoordinates)
    assert "time" in pe.coords
    assert "time" not in pe.get_initialized().coords


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_sizes(pe):
    assert isinstance(pe.sizes, dict)
    assert len(pe.sizes) > len(pe.get_initialized().sizes)


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_dims(pe):
    assert isinstance(pe.dims, xr.core.utils.Frozen)
    assert len(pe.dims) > len(pe.get_initialized().dims)


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_contains(pe):
    v = list(pe.data_vars)[0]
    not_v = v + "notVar"
    assert isinstance(v in pe, bool)
    assert v in pe
    assert not_v not in pe


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_equals(pe):
    assert isinstance(pe.equals(pe), bool)
    assert pe.equals(pe)
    pe2 = pe.copy(deep=False)
    pe2._datasets["initialized"].init.attrs["comment"] = "should not fail"
    assert pe.equals(pe2)

    pe2 = pe2 + 1
    assert not pe.equals(pe2)


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_identical(pe):
    assert isinstance(pe.identical(pe), bool)
    assert pe.identical(pe)

    v = list(pe.data_vars)[0]
    pe2 = pe.copy(deep=False)
    pe2._datasets["initialized"][v].attrs["comment"] = "should fail"
    assert not pe.identical(pe2)


@pytest.mark.parametrize("pe", pe, ids=pe_ids)
def test_PredictionEnsemble_cf(pe):
    """Test that cf_xarray added metadata."""

    # standard_names
    for k, v in CF_STANDARD_NAMES.items():
        if k in pe.coords:
            assert v == pe.get_initialized().coords[k].attrs["standard_name"]

    # long_names
    for k, v in CF_LONG_NAMES.items():
        if k in pe.coords:
            assert v == pe.get_initialized().coords[k].attrs["long_name"]

    # description
    for k, v in CF_LONG_NAMES.items():
        if k in pe.coords:
            assert len(pe.get_initialized().coords[k].attrs["description"]) > 5


def test_warn_if_chunked_along_init_member_time(
    hindcast_hist_obs_1d, perfectModelEnsemble_initialized_control
):
    """Test that _warn_if_chunked_along_init_member_time warns."""
    he = hindcast_hist_obs_1d
    with pytest.warns(UserWarning, match="is chunked along dimensions"):
        he_chunked = HindcastEnsemble(
            he.get_initialized().chunk({"init": 10})
        ).add_observations(he.get_observations())
        with pytest.raises(
            ValueError, match="pass ``allow_rechunk=True`` in ``dask_gufunc_kwargs``"
        ):
            he_chunked.verify(
                metric="rmse", dim="init", comparison="e2o", alignment="same_inits"
            )

    with pytest.warns(UserWarning, match="is chunked along dimensions"):
        he_chunked = HindcastEnsemble(he.get_initialized()).add_observations(
            he.get_observations().chunk({"time": 10})
        )
        with pytest.raises(
            ValueError, match="pass ``allow_rechunk=True`` in ``dask_gufunc_kwargs``"
        ):
            he_chunked.verify(
                metric="rmse", dim="init", comparison="e2o", alignment="same_inits"
            )

    pm = perfectModelEnsemble_initialized_control
    with pytest.warns(UserWarning, match="is chunked along dimensions"):
        PerfectModelEnsemble(pm.get_initialized().chunk({"init": 10}))



class Error(Exception):
    """Base class for custom exceptions in climpred."""

    pass


class CoordinateError(Error):
    """Exception raised when the input xarray object doesn't have the
    appropriate coordinates."""

    def __init__(self, message):
        self.message = message


class DatasetError(Error):
    """Exception raised when there is any issues related to a dataset."""

    def __init__(self, message):
        self.message = message


class DimensionError(Error):
    """Exception raised when the input xarray object doesn't have the
    appropriate dimensions."""

    def __init__(self, message):
        self.message = message


class KeywordError(Error):
    """Exception raised when the keyword used in the function is not appropriate or
    does not work in the given case."""

    def __init__(self, message):
        self.message = message


class VariableError(Error):
    """Exception raised when the input xarray object doesn't have the
    appropriate variables."""

    def __init__(self, message):
        self.message = message



import os

import numpy as np
import pytest
import xarray as xr

from climpred.constants import CLIMPRED_ENSEMBLE_DIMS
from climpred.preprocessing.shared import (
    load_hindcast,
    rename_SLM_to_climpred_dims,
    rename_to_climpred_dims,
    set_integer_time_axis,
)

on_mistral = False
try:
    host = os.environ["HOSTNAME"]
    for node in ["mlogin", "mistralpp"]:
        if node in host:
            on_mistral = True
            from climpred.preprocessing.mpi import get_path
except KeyError:
    pass

# check for intake_esm to be installed
try:
    import intake
    import intake_esm

    print(intake_esm.__version__)
    intake_esm_loaded = True
except ImportError:
    intake_esm_loaded = False


def preprocess_1var(ds, v="global_primary_production"):
    return ds[v].to_dataset(name=v).squeeze()


@pytest.mark.mistral
@pytest.mark.skipif(not on_mistral, reason="requires to be on mistral.dkrz.de")
@pytest.mark.parametrize(
    "inits,members",
    [(range(1961, 1964), range(3, 6)), (range(1970, 1972), range(1, 3))],
)
def test_load_hindcast(inits, members):
    """Test that `load_hindcast` loads the appropriate files."""
    actual = load_hindcast(
        inits=inits,
        members=members,
        preprocess=preprocess_1var,
        get_path=get_path,
    )
    assert isinstance(actual, xr.Dataset)
    assert (actual["init"].values == inits).all()
    assert (actual["member"].values == members).all()
    assert "global_primary_production" in actual.data_vars
    assert len(actual.data_vars) == 1


@pytest.mark.mistral
@pytest.mark.skipif(not on_mistral, reason="requires to be on mistral.dkrz.de")
@pytest.mark.skipif(not intake_esm_loaded, reason="requires intake_esm to be installed")
def test_climpred_pre_with_intake_esm():
    """Test that `preprocess` including `set_integer_time_axis` enables concatination
    of all hindcast into one xr.object."""
    col_url = "/home/mpim/m300524/intake-esm-datastore/catalogs/mistral-cmip6.json"
    col = intake.open_esm_datastore(col_url)
    # load 2 members for 2 inits from one model
    query = dict(
        experiment_id=["dcppA-hindcast"],
        table_id="Amon",
        member_id=["r1i1p1f1", "r2i1p1f1"],
        dcpp_init_year=[1970, 1971],
        variable_id="tas",
        source_id="MPI-ESM1-2-HR",
    )
    cat = col.search(**query)
    cdf_kwargs = {"chunks": {"time": 12}, "decode_times": False}

    def preprocess(ds):
        # extract tiny spatial and temporal subset
        ds = ds.isel(lon=[50, 51, 52], lat=[50, 51, 52], time=np.arange(12 * 2))
        # make time dim identical
        ds = set_integer_time_axis(ds)
        return ds

    dset_dict = cat.to_dataset_dict(cdf_kwargs=cdf_kwargs, preprocess=preprocess)
    # get first dict value
    ds = dset_dict[list(dset_dict.keys())[0]]
    assert isinstance(ds, xr.Dataset)
    ds = rename_to_climpred_dims(ds)
    # check for all CLIMPRED_DIMS
    for c in CLIMPRED_ENSEMBLE_DIMS:
        assert c in ds.coords
    # check for requested dimsizes
    assert ds.member.size == 2
    assert ds.init.size == 2


def test_rename_SLM(da_SLM):
    """Check that dimensions in input are renamed by rename_SLM_to_climpred_dims."""
    da_renamed = rename_SLM_to_climpred_dims(da_SLM)
    for dim in CLIMPRED_ENSEMBLE_DIMS:
        assert dim in da_renamed.dims
    for dim in da_SLM.dims:
        assert dim not in da_renamed.dims


def test_rename_climpred_dims(da_dcpp):
    """Check that dimensions in input are renamed by rename_to_climpred_dims."""
    da_renamed = rename_to_climpred_dims(da_dcpp)
    for dim in CLIMPRED_ENSEMBLE_DIMS:
        assert dim in da_renamed.dims
    for dim in da_dcpp.dims:
        assert dim not in da_renamed.dims



import dask
import pytest
import xarray as xr
from xarray.testing import assert_allclose

from climpred.bootstrap import dpp_threshold
from climpred.stats import decorrelation_time, dpp

try:
    from climpred.bootstrap import varweighted_mean_period_threshold
    from climpred.stats import varweighted_mean_period

    xrft_loaded = True
except ImportError:
    xrft_loaded = False

ITERATIONS = 2


@pytest.mark.parametrize("chunk", (True, False))
def test_dpp(PM_da_control_3d, chunk):
    """Check for positive diagnostic potential predictability in NA SST."""
    res = dpp(PM_da_control_3d, chunk=chunk)
    assert res.mean() > 0


@pytest.mark.skipif(not xrft_loaded, reason="xrft not installed")
@pytest.mark.parametrize("func", (varweighted_mean_period, decorrelation_time))
def test_potential_predictability_likely(PM_da_control_3d, func):
    """Check for positive diagnostic potential predictability in NA SST."""
    res = func(PM_da_control_3d)
    assert res.mean() > 0


def test_bootstrap_dpp_sig50_similar_dpp(PM_da_control_3d):
    sig = 50
    actual = dpp_threshold(PM_da_control_3d, iterations=ITERATIONS, sig=sig).drop_vars(
        "quantile"
    )
    expected = dpp(PM_da_control_3d)
    xr.testing.assert_allclose(actual, expected, atol=0.5, rtol=0.5)


@pytest.mark.skipif(not xrft_loaded, reason="xrft not installed")
def test_bootstrap_vwmp_sig50_similar_vwmp(PM_da_control_3d):
    sig = 50
    actual = varweighted_mean_period_threshold(
        PM_da_control_3d, iterations=ITERATIONS, sig=sig
    ).drop_vars("quantile")
    expected = varweighted_mean_period(PM_da_control_3d)
    xr.testing.assert_allclose(actual, expected, atol=2, rtol=0.5)


def test_bootstrap_func_multiple_sig_levels(PM_da_control_3d):
    sig = [5, 95]
    actual = dpp_threshold(PM_da_control_3d, iterations=ITERATIONS, sig=sig)
    assert actual["quantile"].size == len(sig)
    assert (actual.isel(quantile=0).values <= actual.isel(quantile=1)).all()


@pytest.mark.skipif(not xrft_loaded, reason="xrft not installed")
@pytest.mark.parametrize("step", [1, 2, -1])
@pytest.mark.parametrize(
    "func",
    [dpp, varweighted_mean_period, decorrelation_time],
)
def test_stats_functions_dask_chunks(PM_da_control_3d, func, step):
    """Check whether selected stats functions be chunked and computed along other
    dim."""
    dim = "time"
    for chunk_dim in PM_da_control_3d.isel({dim: 0}).dims:
        control_chunked = PM_da_control_3d.chunk({chunk_dim: step})
        res_chunked = func(control_chunked, dim=dim)
        res = func(PM_da_control_3d, dim=dim)
        # check for chunks
        assert dask.is_dask_collection(res_chunked)
        assert res_chunked.chunks is not None
        # check for no chunks
        assert not dask.is_dask_collection(res)
        assert res.chunks is None
        # check for identical result
        assert_allclose(res, res_chunked.compute())



import dask
import numpy as np
import xarray as xr

from .checks import has_dims, has_min_len
from .constants import M2M_MEMBER_DIM


def _transpose_and_rechunk_to(new_chunk_ds, ori_chunk_ds):
    """Chunk xr.object `new_chunk_ds` as another xr.object `ori_chunk_ds`.
    This is needed after some operations which reduce chunks to size 1.
    First transpose a to ds.dims then apply ds chunking to a."""
    # supposed to be in .utils but circular imports therefore here
    transpose_kwargs = (
        {"transpose_coords": False} if isinstance(new_chunk_ds, xr.DataArray) else {}
    )
    return new_chunk_ds.transpose(*ori_chunk_ds.dims, **transpose_kwargs).chunk(
        ori_chunk_ds.chunks
    )


def _display_comparison_metadata(self):
    summary = "----- Comparison metadata -----\n"
    summary += f"Name: {self.name}\n"
    # probabilistic or only deterministic
    if not self.probabilistic:
        summary += "Kind: deterministic\n"
    else:
        summary += "Kind: deterministic and probabilistic\n"
    summary += f"long_name: {self.long_name}\n"
    # doc
    summary += f"Function: {self.function.__doc__}\n"
    return summary


class Comparison:
    """Master class for all comparisons."""

    def __init__(
        self,
        name,
        function,
        hindcast,
        probabilistic,
        long_name=None,
        aliases=None,
    ):
        """Comparison initialization.

        Args:
            name (str): name of comparison.
            function (function): comparison function.
            hindcast (bool): Can comparison be used in `compute_hindcast`?
                `False` means `compute_perfect_model`
            probabilistic (bool): Can this comparison be used for probabilistic
                metrics also? Probabilistic metrics require multiple forecasts.
                `False` means that comparison is only deterministic.
                `True` means that comparison can be used both deterministic and
                probabilistic.
            long_name (str, optional): longname of comparison. Defaults to None.
            aliases (list of str, optional): Allowed aliases for this comparison.
                Defaults to ``None``.

        Returns:
            comparison: comparison class Comparison.

        """
        self.name = name
        self.function = function
        self.hindcast = hindcast
        self.probabilistic = probabilistic
        self.long_name = long_name
        self.aliases = aliases

    def __repr__(self):
        """Show metadata of comparison class."""
        return _display_comparison_metadata(self)


# --------------------------------------------#
# PERFECT-MODEL COMPARISONS
# --------------------------------------------#


def _m2m(ds, metric=None):
    """Compare all members to all others in turn while leaving out the verification
    ``member``.

    Args:
        ds (xarray object): xr.Dataset/xr.DataArray with ``member`` dimension.
        metric (Metric):
            If deterministic, forecast and reference have ``member`` dim.
            If probabilistic, only forecast has ``member`` dim.

    Returns:
        xr.object: forecast, reference.
    """
    reference_list = []
    forecast_list = []
    for m in ds.member.values:
        forecast = ds.drop_sel(member=m)
        # set incrementing members to avoid nans from broadcasting
        forecast["member"] = np.arange(1, 1 + forecast.member.size)
        reference = ds.sel(member=m, drop=True)
        # Tiles the singular "reference" member to compare directly to all other members
        if not metric.probabilistic:
            forecast, reference = xr.broadcast(forecast, reference)
        reference_list.append(reference)
        forecast_list.append(forecast)
    reference = xr.concat(reference_list, M2M_MEMBER_DIM)
    forecast = xr.concat(forecast_list, M2M_MEMBER_DIM)
    reference[M2M_MEMBER_DIM] = np.arange(reference[M2M_MEMBER_DIM].size)
    forecast[M2M_MEMBER_DIM] = np.arange(forecast[M2M_MEMBER_DIM].size)
    return forecast, reference


__m2m = Comparison(
    name="m2m",
    function=_m2m,
    hindcast=False,
    probabilistic=True,
    long_name="Comparison of all forecasts vs. all other members as verification",
)


def _m2e(ds, metric=None):
    """
    Compare all members to ensemble mean while leaving out the reference in
     ensemble mean.

    Args:
        ds (xarray object): xr.Dataset/xr.DataArray with member and ensemble
                            dimension.
        metric (Metric): needed for probabilistic metrics.
                      therefore useless in m2e comparison,
                      but expected by internal API.

    Returns:
        xr.object: forecast, reference.
    """
    reference_list = []
    forecast_list = []
    M2E_COMPARISON_DIM = "member"
    for m in ds.member.values:
        forecast = ds.drop_sel(member=m).mean("member")
        reference = ds.sel(member=m, drop=True)
        forecast_list.append(forecast)
        reference_list.append(reference)
    reference = xr.concat(reference_list, M2E_COMPARISON_DIM)
    forecast = xr.concat(forecast_list, M2E_COMPARISON_DIM)
    forecast[M2E_COMPARISON_DIM] = np.arange(forecast[M2E_COMPARISON_DIM].size)
    reference[M2E_COMPARISON_DIM] = np.arange(reference[M2E_COMPARISON_DIM].size)
    if dask.is_dask_collection(forecast):
        forecast = _transpose_and_rechunk_to(forecast, ds)
        reference = _transpose_and_rechunk_to(reference, ds)
    return forecast, reference


__m2e = Comparison(
    name="m2e",
    function=_m2e,
    hindcast=False,
    probabilistic=False,
    long_name="Comparison of all members as verification vs. the ensemble mean"
    "forecast",
)


def _m2c(ds, metric=None):
    """
    Compare all other member forecasts to a single member verification, which is the
    first member.
    If the initialized dataset is concatinated in a way that the first member
    is taken from the control simulation, this compares all other member forecasts
    to the control simulation.

    Args:
        ds (xarray object): xr.Dataset/xr.DataArray with member and ensemble
                            dimension.
        metric (Metric): if deterministic, forecast and reference both have member dim
                      if probabilistic, only forecast has member dim

    Returns:
        xr.object: forecast, reference.
    """
    control_member = ds.member.values[0]
    reference = ds.sel(member=control_member, drop=True)
    # drop the member being reference
    forecast = ds.drop_sel(member=control_member)
    if not metric.probabilistic:
        forecast, reference = xr.broadcast(forecast, reference)
    return forecast, reference


__m2c = Comparison(
    name="m2c",
    function=_m2c,
    hindcast=False,
    probabilistic=True,
    long_name="Comparison of multiple forecasts vs. control verification",
)


def _e2c(ds, metric=None):
    """
    Compare ensemble mean forecast to single member verification.
    If the initialized dataset is concatinated in a way that the first member
    is taken from the control simulation, this compares the member mean of all
    other member forecasts to the control simulation.

    Args:
        ds (xarray object): xr.Dataset/xr.DataArray with member and ensemble
                            dimension.
        metric (Metric): needed for probabilistic metrics.
                      therefore useless in e2c comparison,
                      but expected by internal API.

    Returns:
        xr.object: forecast, reference.
    """
    control_member = ds.member.values[0]
    reference = ds.sel(member=control_member, drop=True)
    ds = ds.drop_sel(member=control_member)
    forecast = ds.mean("member")
    return forecast, reference


__e2c = Comparison(
    name="e2c",
    function=_e2c,
    hindcast=False,
    probabilistic=False,
    long_name="Comparison of the ensemble mean forecast vs. control as verification",
)


# --------------------------------------------#
# HINDCAST COMPARISONS
# --------------------------------------------#
def _e2o(hind, verif, metric=None):
    """Compare the ensemble mean forecast to the verification data for a
    ``HindcastEnsemble`` setup.

    Args:
        hind (xarray object): Hindcast with optional ``member`` dimension.
        verif (xarray object): Verification data.
        metric (Metric): needed for probabilistic metrics.
                      therefore useless in ``e2o`` comparison,
                      but expected by internal API.

    Returns:
        xr.object: forecast, verif.
    """
    if "member" in hind.dims:
        forecast = hind.mean("member")
    else:
        forecast = hind
    return forecast, verif


__e2o = Comparison(
    name="e2o",
    function=_e2o,
    hindcast=True,
    probabilistic=False,
    long_name="Verify the ensemble mean against the verification data",
    aliases=["e2r"],
)


def _m2o(hind, verif, metric=None):
    """Compares each ensemble member individually to the verification data for a
    ``HindcastEnsemble`` setup.

    Args:
        hind (xarray object): Hindcast with ``member`` dimension.
        verif (xarray object): Verification data.
        metric (Metric):
            If deterministic, forecast and verif both have ``member`` dim;
            If probabilistic, only forecast has ``member`` dim.

    Returns:
        xr.object: forecast, verif.
    """
    # check that this contains more than one member
    has_dims(hind, "member", "decadal prediction ensemble")
    has_min_len(hind["member"], 1, "decadal prediction ensemble member")
    forecast = hind
    if not metric.probabilistic and "member" not in verif.dims:
        forecast, verif = xr.broadcast(
            forecast, verif, exclude=["time", "init", "lead"]
        )
    return forecast, verif


__m2o = Comparison(
    name="m2o",
    function=_m2o,
    hindcast=True,
    probabilistic=True,
    long_name="Verify each individual forecast member against the verification data.",
    aliases=["m2r"],
)


__ALL_COMPARISONS__ = [__m2m, __m2e, __m2c, __e2c, __e2o, __m2o]

COMPARISON_ALIASES = dict()
for c in __ALL_COMPARISONS__:
    if c.aliases is not None:
        for a in c.aliases:
            COMPARISON_ALIASES[a] = c.name

# Which comparisons work with which set of metrics.
# ['e2o', 'm2o']
HINDCAST_COMPARISONS = [c.name for c in __ALL_COMPARISONS__ if c.hindcast]
# ['m2c', 'e2c', 'm2m', 'm2e']
PM_COMPARISONS = [c.name for c in __ALL_COMPARISONS__ if not c.hindcast]
ALL_COMPARISONS = HINDCAST_COMPARISONS + PM_COMPARISONS
# ['m2c', 'm2m']
PROBABILISTIC_PM_COMPARISONS = [
    c.name for c in __ALL_COMPARISONS__ if (not c.hindcast and c.probabilistic)
]
NON_PROBABILISTIC_PM_COMPARISONS = [
    c.name for c in __ALL_COMPARISONS__ if (not c.hindcast and not c.probabilistic)
]

# ['m2o']
PROBABILISTIC_HINDCAST_COMPARISONS = [
    c.name for c in __ALL_COMPARISONS__ if (c.hindcast and c.probabilistic)
]
PROBABILISTIC_COMPARISONS = (
    PROBABILISTIC_HINDCAST_COMPARISONS + PROBABILISTIC_PM_COMPARISONS
)
ALL_COMPARISONS = [c.name for c in __ALL_COMPARISONS__]



import numpy as np
import pytest
import xarray as xr

from climpred.horizon import _last_item_cond_true, horizon


@pytest.mark.parametrize("input", ["DataArray", "Dataset"])
@pytest.mark.parametrize("threshold,expected", [(1, 1), (2, 3), (3, 6), (3.5, 6)])
def test_least_item_cond_true(threshold, expected, input):
    """test `last_item_cond_true` on artificial data."""
    ds = xr.DataArray(
        [1, 2, 2, 3, 3, 1, 4], dims="lead", coords={"lead": np.arange(1, 1 + 7)}
    )
    if input == "Dataset":
        ds = ds.to_dataset(name="test")
    cond = ds <= threshold
    actual = _last_item_cond_true(cond, "lead")
    assert actual == expected
    assert type(ds) == type(actual)


def test_horizon_bootstrap_1d(perfectModelEnsemble_initialized_control):
    """test horizon for pm.bootstrap for 1d."""
    bskill = perfectModelEnsemble_initialized_control.bootstrap(
        iterations=201,
        metric="rmse",
        comparison="m2e",
        dim=["member", "init"],
        reference="uninitialized",
    )
    ph = horizon(bskill.sel(results="p", skill="uninitialized") <= 0.05)
    assert ph.tos.attrs["units"] == "years", print(ph.tos)
    assert int(ph.tos) in [4, 5, 6, 7, 8]  # should be 6, testing on the safe side


def test_horizon_3d(hindcast_recon_3d):
    he = hindcast_recon_3d
    # set one grid to all nan mimicking land
    land = he._datasets["initialized"]["SST"][:, :, 0, 0]
    he._datasets["initialized"]["SST"][:, :, 0, 0] = land.where(land == 100000)
    # verify
    skill = he.verify(
        metric="acc",
        comparison="e2o",
        dim="init",
        alignment="maximize",
        reference=["persistence"],
    )

    ph = horizon(skill.sel(skill="initialized") > skill.sel(skill="persistence"))
    assert "variable" not in skill.dims
    assert "variable" not in skill.coords
    # test all nan on land
    assert ph["SST"][0, 0].isnull()
    # test significant everywhere
    assert (ph >= 1).all()
    assert (ph.isel(nlat=-1) == 2).all()


@pytest.mark.parametrize("smooth", [True, False])
def test_horizon_smooth(perfectModelEnsemble_initialized_control, smooth):
    """test horizon for pm.smooth(lead).verify."""
    pm = perfectModelEnsemble_initialized_control
    if smooth:
        pm = pm.smooth(
            {"lead": 2}, how="mean"
        )  # converts lead to '1-2', '2-3', ... after verify
    skill = pm.verify(
        metric="rmse",
        comparison="m2e",
        dim=["member", "init"],
        reference="persistence",
    )
    assert skill.lead.attrs["units"] == "years", print(skill.lead.attrs)
    # initialized better than persistence if RMSE smaller
    cond = skill.sel(skill="initialized", drop=True) < skill.sel(
        skill="persistence", drop=True
    )
    ph = horizon(cond)
    assert ph.tos.attrs["units"] == "years", print("ph.tos.attrs = ", ph.tos.attrs)
    assert ph.tos.values == skill.lead.isel(lead=-1).values


def test_horizon_weird_coords():
    """Test horizon for weird coords."""
    cond = xr.DataArray([True] * 10, dims="lead").to_dataset(name="SST")
    # Change leads to something weird
    cond["lead"] = [0.25, 0.75, 1, 3, 4, 5, 6, 7, 8, 9]
    assert _last_item_cond_true(cond, "lead") == 9.0

    cond[0] = False
    assert _last_item_cond_true(cond, "lead").isnull(), print(cond)

    cond[0] = True
    cond[1] = False
    assert _last_item_cond_true(cond, "lead") == cond.lead[0], print(cond)

    cond["lead"] = np.arange(cond.lead.size)
    cond[0] = True
    cond[1] = False
    assert _last_item_cond_true(cond, "lead") == 0.0, print(cond)



import dask
import numpy as np
import pytest
import xarray as xr

from climpred.bootstrap import bootstrap_perfect_model
from climpred.constants import CLIMPRED_DIMS
from climpred.metrics import DETERMINISTIC_PM_METRICS
from climpred.prediction import compute_perfect_model
from climpred.reference import compute_persistence

xr.set_options(display_style="text")


@pytest.mark.parametrize("metric", ["mse", "pearson_r"])
def test_compute_persistence_lead0_lead1(
    PM_da_initialized_1d, PM_da_initialized_1d_lead0, PM_da_control_1d, metric
):
    """
    Checks that persistence forecast results are identical for a lead 0 and lead 1 setup
    """
    res1 = compute_persistence(
        PM_da_initialized_1d, PM_da_control_1d, metric=metric, alignment="same_inits"
    )
    res2 = compute_persistence(
        PM_da_initialized_1d_lead0,
        PM_da_control_1d,
        metric=metric,
        alignment="same_inits",
    )
    assert (res1.values == res2.values).all()


@pytest.mark.parametrize("metric", ("AnomCorr", "test", "None"))
def test_compute_perfect_model_metric_keyerrors(
    PM_da_initialized_1d, PM_da_control_1d, metric
):
    """
    Checks that wrong metric names get caught.
    """
    with pytest.raises(KeyError) as excinfo:
        compute_perfect_model(
            PM_da_initialized_1d,
            PM_da_control_1d,
            comparison="e2c",
            metric=metric,
        )
    assert "Specify metric from" in str(excinfo.value)


@pytest.mark.parametrize("comparison", ("ensemblemean", "test", "None"))
def test_compute_perfect_model_comparison_keyerrors(
    PM_da_initialized_1d, PM_da_control_1d, comparison
):
    """
    Checks that wrong comparison names get caught.
    """
    with pytest.raises(KeyError) as excinfo:
        compute_perfect_model(
            PM_da_initialized_1d,
            PM_da_control_1d,
            comparison=comparison,
            metric="mse",
        )
    assert "Specify comparison from" in str(excinfo.value)



import dask
import numpy as np
import pytest

from climpred.bootstrap import bootstrap_hindcast
from climpred.comparisons import HINDCAST_COMPARISONS
from climpred.constants import CLIMPRED_DIMS
from climpred.prediction import compute_hindcast
from climpred.reference import compute_persistence

ITERATIONS = 2


def test_compute_hindcast_lead0_lead1(
    hind_ds_initialized_1d, hind_ds_initialized_1d_lead0, reconstruction_ds_1d
):
    """
    Checks that compute hindcast returns the same results with a lead-0 and lead-1
    framework.
    """
    res1 = compute_hindcast(
        hind_ds_initialized_1d,
        reconstruction_ds_1d,
        metric="rmse",
        comparison="e2o",
    )
    res2 = compute_hindcast(
        hind_ds_initialized_1d_lead0,
        reconstruction_ds_1d,
        metric="rmse",
        comparison="e2o",
    )
    assert (res1.SST.values == res2.SST.values).all()


def test_persistence_lead0_lead1(
    hind_ds_initialized_1d, hind_ds_initialized_1d_lead0, reconstruction_ds_1d
):
    """
    Checks that compute persistence returns the same results with a lead-0 and lead-1
    framework.
    """
    res1 = compute_persistence(
        hind_ds_initialized_1d, reconstruction_ds_1d, metric="rmse"
    )
    res2 = compute_persistence(
        hind_ds_initialized_1d_lead0, reconstruction_ds_1d, metric="rmse"
    )
    assert (res1.SST.values == res2.SST.values).all()


@pytest.mark.parametrize("metric", ("AnomCorr", "test", "None"))
def test_compute_hindcast_metric_keyerrors(
    hind_ds_initialized_1d, reconstruction_ds_1d, metric
):
    """
    Checks that wrong metric names get caught.
    """
    with pytest.raises(KeyError) as excinfo:
        compute_hindcast(
            hind_ds_initialized_1d,
            reconstruction_ds_1d,
            comparison="e2o",
            metric=metric,
        )
    assert "Specify metric from" in str(excinfo.value)


@pytest.mark.parametrize("comparison", ("ensemblemean", "test", "None"))
def test_compute_hindcast_comparison_keyerrors(
    hind_ds_initialized_1d, reconstruction_ds_1d, comparison
):
    """
    Checks that wrong comparison names get caught.
    """
    with pytest.raises(KeyError) as excinfo:
        compute_hindcast(
            hind_ds_initialized_1d,
            reconstruction_ds_1d,
            comparison=comparison,
            metric="mse",
        )
    assert "Specify comparison from" in str(excinfo.value)


@pytest.mark.parametrize("metric", ("rmse", "pearson_r"))
def test_compute_hindcast_dask_spatial(
    hind_da_initialized_3d, reconstruction_da_3d, metric
):
    """Chunking along spatial dims."""
    # chunk over dims in both
    for dim in hind_da_initialized_3d.dims:
        if dim in reconstruction_da_3d.dims:
            step = 5
            res_chunked = compute_hindcast(
                hind_da_initialized_3d.chunk({dim: step}),
                reconstruction_da_3d.chunk({dim: step}),
                comparison="e2o",
                metric=metric,
            )
            # check for chunks
            assert dask.is_dask_collection(res_chunked)
            assert res_chunked.chunks is not None


def test_compute_hindcast_CESM_3D_keep_coords(
    hind_da_initialized_3d, reconstruction_da_3d
):
    """Test that no coords are lost in compute_hindcast with the CESM sample data."""
    s = compute_hindcast(hind_da_initialized_3d, reconstruction_da_3d)
    for c in hind_da_initialized_3d.drop("init").coords:
        assert c in s.coords


def test_HindcastEnsemble_keeps_lead_units(hindcast_hist_obs_1d):
    """Test that lead units is kept in bootstrap."""
    sig = 95
    actual = hindcast_hist_obs_1d.bootstrap(
        metric="mse",
        iterations=ITERATIONS,
        comparison="e2o",
        sig=sig,
        dim="init",
        alignment="same_verif",
    )
    assert actual.lead.attrs["units"] == "years"



import glob as _glob
import os as _os


def get_path(
    dir_base_experiment="/work/bm1124/m300086/CMIP6/experiments",
    member=1,
    init=1960,
    model="hamocc",
    output_stream="monitoring_ym",
    timestr="*1231",
    ending="nc",
):  # “pragma: no cover”
    """Get the path of a file for MPI-ESM standard output file names and directory.

    Args:
        dir_base_experiment (str): Path of experiments folder. Defaults to
            "/work/bm1124/m300086/CMIP6/experiments".
        member (int): member label. Defaults to 1.
        init (int): initialization label. Typically year. Defaults to 1960.
        model (str): submodel name. Defaults to "hamocc".
            Allowed: ['echam6', 'jsbach', 'mpiom', 'hamocc'].
        output_stream (str): output_stream name. Defaults to "monitoring_ym".
            Allowed: ['data_2d_mm', 'data_3d_ym', 'BOT_mm', ...]
        timestr (str): timestr likely including *. Defaults to "*1231".
        ending (str): ending indicating file format. Defaults to "nc".
            Allowed: ['nc', 'grb'].

    Returns:
        str: path of requested file(s)

    """
    # get experiment_id
    dirs = _os.listdir(dir_base_experiment)
    experiment_id = [
        x for x in dirs if (f"{init}" in x and "r" + str(member) + "i" in x)
    ]
    assert len(experiment_id) == 1
    experiment_id = experiment_id[0]
    dir_outdata = f"{dir_base_experiment}/{experiment_id}/outdata/{model}"
    path = f"{dir_outdata}/{experiment_id}_{model}_{output_stream}_{timestr}.{ending}"
    if _os.path.exists(_glob.glob(path)[0]):
        return path
    else:
        raise ValueError(f"Path not found or no access: {path}")
